import io
import os
import shutil
import tempfile

import paramiko
from django.utils.encoding import force_str
from djblets.siteconfig.models import SiteConfiguration
from djblets.util.decorators import cached_property

from reviewboard.ssh.client import SSHClient
from reviewboard.ssh.errors import UnsupportedSSHKeyError
from reviewboard.ssh.storage import (FileSSHStorage,
                                     get_ssh_storage_backend_path,
                                     set_ssh_storage_backend_path)
from reviewboard.ssh.utils import humanize_key
from reviewboard.testing.testcase import TestCase


rsa_key_blob = """-----BEGIN RSA PRIVATE KEY-----
MIICXQIBAAKBgQCsElbDaXtsLctXBQVu8N55ptQ3s1IDBP1nqL/0J3+L70DMjRXa
tVB9uPeZOPDJrWgu7Gn2k48oRkGdl+9+WtnaNvgb6jC9TU4gNXKtBeq/Q/NQgtLs
jBnhczMC90PnM+Bvq6TyDaufXqYP7w8Xk1TsW7nz58HMsIPOEA8Ajx+3PwIDAQAB
AoGAXgjbn4j6qSDRmemlkX5Spnq0SQhXTk0gytBermgTfP6wE9kaU1548Wvu665B
cIWyhMowEk+LkX/rhdstR4kQuhkgGtkO78YLjqmHPHuMYRn4Ea/1xdSYA1qOnLWR
GNbnnvYY9/YR5KhsmFbuG5wfA2V0Bw3ULm02jgGuCV7Y5okCQQDgN4md1qXmkO9S
XgfftE1r4ByqFwWzzFRTAFEFN2jULUwmM3B+L+1MUGjuKBk/tjfdv7QBPRJoO6xz
peG00nHNAkEAxHaIyIaaK9ajrke+tiSDZYs9HCnYHiVH2+3hg1vTHIrgO8VkjA93
A40Qaol+7dKzsC5TPll3k2uGnY+lo/RxOwJAI+RgEDc7KXSMCvhodEQNnLYsgIHc
9NJBsWO8lIQxML3rkbXsTRbo+q1ojq82k39c5A97BjO7jZn32i90uRhzBQJBALcQ
KHaJjeDpeM1thtRcA5+79a5ngzzbyjCxYSAwkO+YrEalsQIdau2BJVnQUtiyK8Mv
91syrIxOdjoc3uB+Zn8CQQCpvbEXIU/76MH/yDmgOk4+R8qo/yU6cgn7PTCWzGL7
SK+fSBGKFq+n2FxQIt9OWswQ+wbvq9jmJmLCGxuUSMPu
-----END RSA PRIVATE KEY-----
"""

dsa_key_blob = """-----BEGIN DSA PRIVATE KEY-----
MIIBugIBAAKBgQDddn3Hr3guZXLlmRLlneT0HSUa3gx3dYVCMr/b7UXu7gMxG919
C6Tzjk300tgxDpTnmq1OVwoQA44tIFYlxvw9KnxttnPe+Ny7nocGDApBXMLfaZLN
QbAlsBxTEVPB6CxtF9srVs3SXNbQddGI/PidEK00Fe1jwNnv0aC43LCFFwIVAM/d
qNnjATC1+ub/4dwnbO4sL2zlAoGAVM/g9ePoFxdldGKh40SaNHjkSw9GMo72HioD
KkSBNJ2Es/8ppX6Wkgi3WWZNsMruTTnVyWPqPIPpt58yqyMYtqSVVmoK7ihyxbxW
dUtG9rrNwo9/OqfvUxGFYE0suBnNR29lKKlWT+Sk5Cjd+5BpGZ6ptaxgvkYDFkyX
JrWBXzUCgYA0u51vP+h9InIxYxAr64Y72rungv/2Y409vvEbnBDK42na8SJ4fNZF
CUa4Y8KQ8bUaKyBbiXz/r+zbzA7D5kxsdBMeUmHjQhMIGiMxvGfPLw/9jWR2pcFH
DPCGtVEaccnAOCgOEfgRGq5MG/i0YCFj7AIdLQchGiUDVPJNFK8KNwIUUDs/Ac/t
NnIFhSieTpeXxmozkks=
-----END DSA PRIVATE KEY-----
"""


class TestKeys(object):
    """Keys used for unit tests.

    This is used to access keys across any and all unit tests that need them,
    in a way that reduces overhead by constructing each key only once and
    only on first access, caching it for future lookups.
    """

    @cached_property
    def rsa_key(self):
        """A stable RSA key for testing."""
        return paramiko.RSAKey.from_private_key(io.StringIO(rsa_key_blob))

    @cached_property
    def dsa_key(self):
        """A stable DSA key for testing."""
        return paramiko.DSSKey.from_private_key(io.StringIO(dsa_key_blob))

    @cached_property
    def rsa_key_b64(self):
        """Base64 encoding for the RSA key."""
        return test_keys.rsa_key.get_base64()

    @cached_property
    def dsa_key_b64(self):
        """Base64 encoding for the DSA key."""
        return test_keys.dsa_key.get_base64()


test_keys = TestKeys()


class SSHTestCase(TestCase):
    def setUp(self):
        super(SSHTestCase, self).setUp()

        self.old_home = os.getenv('HOME')
        self.tempdir = None
        os.environ[str('RBSSH_ALLOW_AGENT')] = str('0')
        FileSSHStorage._ssh_dir = None

    def tearDown(self):
        super(SSHTestCase, self).tearDown()

        self._set_home(self.old_home)

        if self.tempdir:
            shutil.rmtree(self.tempdir)

    @property
    def key1(self):
        """Legacy alias for TestKeys.rsa_key."""
        return test_keys.rsa_key

    @property
    def key2(self):
        """Legacy alias for TestKeys.dsa_key."""
        return test_keys.dsa_key

    @property
    def key1_b64(self):
        """Legacy alias for TestKeys.rsa_key_b64."""
        return test_keys.rsa_key_b64

    @property
    def key2_b64(self):
        """Legacy alias for TestKeys.dsa_key_b64."""
        return test_keys.dsa_key_b64

    def _set_home(self, homedir):
        os.environ[str('HOME')] = force_str(homedir)


class FileSSHStorageTests(SSHTestCase):
    """Unit tests for FileSSHStorage."""
    def setUp(self):
        super(FileSSHStorageTests, self).setUp()

        self.tempdir = tempfile.mkdtemp(prefix='rb-tests-home-')
        self._set_home(self.tempdir)

    def test_get_ssh_dir_with_dot_ssh(self):
        """Testing FileSSHStorage.get_ssh_dir with ~/.ssh"""
        sshdir = os.path.join(self.tempdir, '.ssh')

        storage = FileSSHStorage()
        self.assertEqual(storage.get_ssh_dir(), sshdir)

    def test_get_ssh_dir_with_ssh(self):
        """Testing FileSSHStorage.get_ssh_dir with ~/ssh"""
        sshdir = os.path.join(self.tempdir, 'ssh')
        os.mkdir(sshdir, 0o700)

        storage = FileSSHStorage()
        self.assertEqual(storage.get_ssh_dir(), sshdir)

    def test_get_ssh_dir_with_dot_ssh_and_localsite(self):
        """Testing FileSSHStorage.get_ssh_dir with ~/.ssh and localsite"""
        sshdir = os.path.join(self.tempdir, '.ssh', 'site-1')

        storage = FileSSHStorage(namespace='site-1')
        self.assertEqual(storage.get_ssh_dir(), sshdir)

    def test_get_ssh_dir_with_ssh_and_localsite(self):
        """Testing FileSSHStorage.get_ssh_dir with ~/ssh and localsite"""
        sshdir = os.path.join(self.tempdir, 'ssh')
        os.mkdir(sshdir, 0o700)
        sshdir = os.path.join(sshdir, 'site-1')

        storage = FileSSHStorage(namespace='site-1')
        self.assertEqual(storage.get_ssh_dir(), sshdir)

    def test_write_user_key_unsupported(self):
        """Testing FileSSHStorage.write_user_key with unsupported key type"""
        class FakeKey(object):
            pass

        storage = FileSSHStorage()
        self.assertRaises(UnsupportedSSHKeyError,
                          lambda: storage.write_user_key(FakeKey()))

    def test_read_host_keys(self):
        """Testing FileSSHStorage.read_host_keys"""
        storage = FileSSHStorage()
        storage.ensure_ssh_dir()

        line1 = 'host1 ssh-rsa %s' % test_keys.rsa_key_b64
        line2 = 'host2 ssh-dss %s' % test_keys.dsa_key_b64

        filename = storage.get_host_keys_filename()

        with open(filename, 'w') as fp:
            fp.write('%s\n' % line1)
            fp.write('\n')
            fp.write('# foo\n')
            fp.write('%s  \n' % line2)

        lines = storage.read_host_keys()
        self.assertEqual(len(lines), 2)
        self.assertEqual(lines[0], line1)
        self.assertEqual(lines[1], line2)

    def test_add_host_key(self):
        """Testing FileSSHStorage.add_host_key"""
        storage = FileSSHStorage()
        storage.add_host_key('host1', test_keys.rsa_key)

        filename = storage.get_host_keys_filename()

        with open(filename, 'r') as fp:
            lines = fp.readlines()

        self.assertEqual(len(lines), 1)
        self.assertEqual(lines[0],
                         'host1 ssh-rsa %s\n' % test_keys.rsa_key_b64)

    def test_replace_host_key(self):
        """Testing FileSSHStorage.replace_host_key"""
        storage = FileSSHStorage()
        storage.add_host_key('host1', test_keys.rsa_key)
        storage.replace_host_key('host1', test_keys.rsa_key,
                                 test_keys.dsa_key)

        filename = storage.get_host_keys_filename()

        with open(filename, 'r') as fp:
            lines = fp.readlines()

        self.assertEqual(len(lines), 1)
        self.assertEqual(lines[0],
                         'host1 ssh-dss %s\n' % test_keys.dsa_key_b64)

    def test_replace_host_key_no_known_hosts(self):
        """Testing FileSSHStorage.replace_host_key with no known hosts file"""
        storage = FileSSHStorage()
        storage.replace_host_key('host1', test_keys.rsa_key, test_keys.dsa_key)

        filename = storage.get_host_keys_filename()

        with open(filename, 'r') as fp:
            lines = fp.readlines()

        self.assertEqual(len(lines), 1)
        self.assertEqual(lines[0],
                         'host1 ssh-dss %s\n' % test_keys.dsa_key_b64)


class SSHClientTests(SSHTestCase):
    """Unit tests for SSHClient."""
    def setUp(self):
        super(SSHClientTests, self).setUp()

        self.tempdir = tempfile.mkdtemp(prefix='rb-tests-home-')

    def test_generate_user_key(self, namespace=None):
        """Testing SSHClient.generate_user_key"""
        self._set_home(self.tempdir)

        client = SSHClient(namespace=namespace)
        key = client.generate_user_key(bits=1024)
        key_file = os.path.join(client.storage.get_ssh_dir(), 'id_rsa')
        self.assertTrue(os.path.exists(key_file))
        self.assertEqual(client.get_user_key(), key)

    def test_generate_user_key_with_localsite(self):
        """Testing SSHClient.generate_user_key with localsite"""
        self.test_generate_user_key('site-1')

    def test_delete_user_key(self, namespace=None):
        """Testing SSHClient.delete_user_key"""
        self._set_home(self.tempdir)

        client = SSHClient(namespace=namespace)
        client.import_user_key(test_keys.rsa_key)

        key_file = os.path.join(client.storage.get_ssh_dir(), 'id_rsa')
        self.assertTrue(os.path.exists(key_file))
        self.assertEqual(client.get_user_key(), test_keys.rsa_key)

        client.delete_user_key()
        self.assertFalse(os.path.exists(key_file))

    def test_delete_user_key_with_localsite(self):
        """Testing SSHClient.delete_user_key with localsite"""
        self.test_delete_user_key('site-1')

    def test_add_host_key(self, namespace=None):
        """Testing SSHClient.add_host_key"""
        self._set_home(self.tempdir)
        client = SSHClient(namespace=namespace)

        client.add_host_key('example.com', test_keys.rsa_key)

        known_hosts_file = client.storage.get_host_keys_filename()
        self.assertTrue(os.path.exists(known_hosts_file))

        with open(known_hosts_file, 'r') as f:
            lines = f.readlines()

        self.assertEqual(len(lines), 1)
        self.assertEqual(lines[0].split(),
                         ['example.com', test_keys.rsa_key.get_name(),
                          test_keys.rsa_key_b64])

    def test_add_host_key_with_localsite(self):
        """Testing SSHClient.add_host_key with localsite"""
        self.test_add_host_key('site-1')

    def test_replace_host_key(self, namespace=None):
        """Testing SSHClient.replace_host_key"""
        self._set_home(self.tempdir)
        client = SSHClient(namespace=namespace)

        client.add_host_key('example.com', test_keys.rsa_key)
        client.replace_host_key('example.com', test_keys.rsa_key,
                                test_keys.dsa_key)

        known_hosts_file = client.storage.get_host_keys_filename()
        self.assertTrue(os.path.exists(known_hosts_file))

        with open(known_hosts_file, 'r') as f:
            lines = f.readlines()

        self.assertEqual(len(lines), 1)
        self.assertEqual(lines[0].split(),
                         ['example.com', test_keys.dsa_key.get_name(),
                          test_keys.dsa_key_b64])

    def test_replace_host_key_with_localsite(self):
        """Testing SSHClient.replace_host_key with localsite"""
        self.test_replace_host_key('site-1')

    def test_import_user_key(self, namespace=None):
        """Testing SSHClient.import_user_key"""
        self._set_home(self.tempdir)
        client = SSHClient(namespace=namespace)

        client.import_user_key(test_keys.rsa_key)
        self.assertEqual(client.get_user_key(), test_keys.rsa_key)

    def test_import_user_key_with_localsite(self):
        """Testing SSHClient.import_user_key with localsite"""
        self.test_import_user_key('site-1')


class SettingsTests(TestCase):
    """Unit tests for the SSH storage backend settings functions."""

    def setUp(self):
        super(SettingsTests, self).setUp()

        self.siteconfig = SiteConfiguration.objects.get_current()

    def test_set_ssh_storage_backend_path(self):
        """Testing set_ssh_storage_backend_path"""
        set_ssh_storage_backend_path('foo.bar.FooStorage')

        self.assertEqual(self.siteconfig.get('ssh_storage_backend'),
                         'foo.bar.FooStorage')

    def test_get_ssh_storage_backend_path(self):
        """Testing set_ssh_storage_backend_path"""
        self.siteconfig.set('ssh_storage_backend', 'foo.bar.BarStorage')

        self.assertEqual(get_ssh_storage_backend_path(),
                         'foo.bar.BarStorage')


class UtilsTests(SSHTestCase):
    """Unit tests for reviewboard.ssh.utils."""

    def test_humanize_key_with_rsa_key(self):
        """Testing humanize_key with RSA key"""
        humanized = humanize_key(test_keys.rsa_key)
        self.assertIsInstance(humanized, str)
        self.assertEqual(humanized,
                         '76:ec:40:bd:69:9e:b1:e4:47:a9:e3:74:82:ec:0c:0f')

    def test_humanize_key_with_dsa_key(self):
        """Testing humanize_key with DSA key"""
        humanized = humanize_key(test_keys.dsa_key)
        self.assertIsInstance(humanized, str)
        self.assertEqual(humanized,
                         '62:4b:7f:b0:94:57:e2:bb:e7:d8:a4:88:88:c6:10:38')



"""View for the New Review Request page."""

import logging
from typing import Any, Dict

from django.views.generic.base import TemplateView

from reviewboard.accounts.mixins import (LoginRequiredViewMixin,
                                         UserProfileRequiredViewMixin)
from reviewboard.admin.mixins import CheckReadOnlyViewMixin
from reviewboard.scmtools.models import Repository
from reviewboard.site.mixins import CheckLocalSiteAccessViewMixin


logger = logging.getLogger(__name__)


class NewReviewRequestView(LoginRequiredViewMixin,
                           CheckLocalSiteAccessViewMixin,
                           UserProfileRequiredViewMixin,
                           CheckReadOnlyViewMixin,
                           TemplateView):
    """View for the New Review Request page.

    This provides the user with a UI consisting of all their repositories,
    allowing them to manually upload a diff against the repository or,
    depending on the repository's capabilities, to browse for an existing
    commit to post.
    """

    template_name = 'reviews/new_review_request.html'

    def get_context_data(
        self,
        **kwargs,
    ) -> Dict[str, Any]:
        """Return data for the template.

        This will return information on each repository shown on the page.

        Args:
            **kwargs (dict):
                Additional keyword arguments passed to the view.

        Returns:
            dict:
            Context data for the template.
        """
        local_site = self.local_site

        if local_site:
            local_site_prefix = 's/%s/' % local_site.name
        else:
            local_site_prefix = ''

        valid_repos = []

        repos = Repository.objects.accessible(self.request.user,
                                              local_site=local_site)

        for repo in repos.order_by('name')[:25]:
            try:
                valid_repos.append({
                    'id': repo.pk,
                    'name': repo.name,
                    'scmtoolName': repo.scmtool_class.name,
                    'localSitePrefix': local_site_prefix,
                    'supportsPostCommit': repo.supports_post_commit,
                    'requiresChangeNumber': repo.supports_pending_changesets,
                    'requiresBasedir': not repo.diffs_use_absolute_paths,
                    'filesOnly': False,
                })
            except Exception:
                logger.exception(
                    'Error loading information for repository "%s" (ID %d) '
                    'for the New Review Request page.',
                    repo.name, repo.pk)

        return {
            'page_model_attrs': {
                'repositories': valid_repos,
                'localSitePrefix': local_site_prefix,
            }
        }



"""Settings used to manage the processing and render of diffs.

Version Added:
    5.0.2
"""

from __future__ import annotations

import json
from dataclasses import dataclass, fields as dataclass_fields
from hashlib import sha256
from typing import Any, ClassVar, Dict, List, Optional, TYPE_CHECKING, cast

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.utils.functional import cached_property
from djblets.siteconfig.models import SiteConfiguration

if TYPE_CHECKING:
    from django.http import HttpRequest
    from typelets.json import JSONDict

    from reviewboard.site.models import LocalSite


@dataclass
class DiffSettings:
    """Settings used to render a diff.

    These settings represent a combination of site configuration settings for
    the diff viewer and user-configured settings. They're used to control how
    the diff viewer displays diffs, covering syntax highlighting options,
    code safety checks, whitespace handling, and more.

    Version Added:
        5.0.2
    """

    #: The default for tabstop widths.
    #:
    #: Version Added:
    #:     7.0.4
    DEFAULT_TAB_SIZE: ClassVar[int] = 8

    ######################
    # Instance variables #
    ######################

    #: A mapping of code safety checker IDs to configurations.
    #:
    #: Type:
    #:     dict
    code_safety_configs: dict[str, dict[str, Any]]

    #: The number of lines of context to show around modifications in a chunk.
    #:
    #: Type:
    #:     int
    context_num_lines: int

    #: A custom mapping of file extensions to Pygments lexers.
    #:
    #: Type:
    #:     dict
    custom_pygments_lexers: dict[str, str]

    #: A list of file globs for which legacy whitespace rules should be used.
    #:
    #: Any file matching a pattern in this list will treat all whitespace as
    #: modifications to lines. Smart indentation handling and other features
    #: may behave differently.
    #:
    #: This is considered a legacy feature.
    #:
    #: Type:
    #:     list of str
    include_space_patterns: list[str]

    #: Whether to filter upstream changes out of an interdiff.
    #:
    #: When enabled, this will filter out upstream changes that may be
    #: introduced between two revisions of any given file in an interdiff.
    #:
    #: This defaults to ``True``.
    #:
    #: Version Added:
    #:     8.0
    #:
    #: Type:
    #:     bool
    interdiff_filtering: bool

    #: The number of files to include in each page of a diff.
    #:
    #: Type:
    #:     int
    paginate_by: int

    #: The maximum number of orphans to include on the last page of a diff.
    #:
    #: Orphans are extra files that would make up less than a full page. If
    #: there aren't more than this number of orphans, they'll be rolled up
    #: into the final page of the diff viewer.
    #:
    #: Type:
    #:     int
    paginate_orphans: int

    #: Whether to enable syntax highlighting.
    #:
    #: Type:
    #:     bool
    syntax_highlighting: bool

    #: The maximum number of lines in a file to allow syntax highlighting.
    #:
    #: If a file has lines beyond this threshold, syntax highlighting will
    #: be forcefully disabled.
    #:
    #: Type:
    #:     int
    syntax_highlighting_threshold: int

    #: The default tabstop width for diffs.
    #:
    #: Version Added:
    #:     7.0.4
    tab_size: int

    @classmethod
    def create(
        cls,
        *,
        user: (User | None) = None,
        local_site: (LocalSite | None) = None,
        request: (HttpRequest | None) = None,
        interdiff_filtering: bool = True,
        syntax_highlighting: (bool | None) = None,
    ) -> DiffSettings:
        """Create diff settings based on the provided arguments.

        This will compute settings based on any HTTP request, Local Site, or
        user provided. The rest will be filled out from the site configuration.

        Syntax highlighting will be enabled if it's both enabled in the site
        configuration and in the user's profile.

        If ``request`` is provided, its associated user and Local Site will
        be used as defaults if ``user`` or ``local_site`` are not provided.

        Version Added:
            8.0:
            Added the ``interdiff_filtering`` argument.

        Args:
            user (django.contrib.auth.models.User, optional):
                The user the settings should pertain to.

            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site that the settings should pertain to.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

            interdiff_filtering (bool, optional):
                Whether to filter upstream changes out of an interdiff.

                When enabled, this will filter out upstream changes that may
                be introduced between two revisions of any given file in an
                interdiff.

                This defaults to ``True``.

                Version Added:
                    8.0

            syntax_highlighting (bool, optional):
                An explicit value for :py:attr:`syntax_highlighting`. If
                provided, the user and site settings will be ignored.

        Returns:
            DiffSettings:
            The settings for rendering the diff.
        """
        siteconfig = SiteConfiguration.objects.get_current()

        # If a request is passed, see if we need to extract information from
        # it to fill in missing attributes.
        if request is not None:
            if user is None and request.user.is_authenticated:
                # Satisfy the type checker.
                assert isinstance(request.user, User)

                user = request.user

            if local_site is None:
                local_site = getattr(request, 'local_site', None)

        # Set up config layers to allow Local Sites to override the global
        # settings. We may eventually want to add repository- and user-level
        # settings as well.
        config_layers: list[JSONDict] = []

        if (local_site and
            (local_site_config := local_site.extra_data.get('siteconfig'))):
            config_layers.append(local_site_config)

        # Figure out the default for syntax highlighting.
        if syntax_highlighting is None:
            syntax_highlighting = cast(
                bool,
                siteconfig.get('diffviewer_syntax_highlighting',
                               layers=config_layers))

            if syntax_highlighting and user and user.is_authenticated:
                # The server enables syntax highlighting. See if the user has
                # enabled or disabled it.
                try:
                    profile = user.get_profile()  # type: ignore
                    syntax_highlighting = profile.syntax_highlighting
                except ObjectDoesNotExist:
                    pass

            assert syntax_highlighting is not None

        tab_size = cast(
            Optional[int],
            siteconfig.get('diffviewer_default_tab_size',
                           layers=config_layers))

        # If the tab size is zero or unset, set to default.
        if not tab_size:
            tab_size = cls.DEFAULT_TAB_SIZE

        return cls(
            code_safety_configs=cast(
                Dict[str, Dict[str, Any]],
                siteconfig.get('code_safety_checkers',
                               layers=config_layers)
            ),
            context_num_lines=cast(
                int,
                siteconfig.get('diffviewer_context_num_lines',
                               layers=config_layers)),
            custom_pygments_lexers=cast(
                Dict[str, str],
                siteconfig.get('diffviewer_custom_pygments_lexers',
                               layers=config_layers)),
            include_space_patterns=cast(
                List[str],
                siteconfig.get('diffviewer_include_space_patterns',
                               layers=config_layers)),
            interdiff_filtering=interdiff_filtering,
            paginate_by=cast(
                int,
                siteconfig.get('diffviewer_paginate_by',
                               layers=config_layers)),
            paginate_orphans=cast(
                int,
                siteconfig.get('diffviewer_paginate_orphans',
                               layers=config_layers)),
            syntax_highlighting=syntax_highlighting,
            syntax_highlighting_threshold=cast(
                int,
                siteconfig.get('diffviewer_syntax_highlighting_threshold',
                               layers=config_layers)),
            tab_size=tab_size,
        )

    @cached_property
    def state_hash(self) -> str:
        """Return a hash of the current settings.

        This can be used as a component in cache keys and ETags to ensure
        that changes to diff settings trigger re-generations of diffs.

        This is calculated only once per instance. It won't take into account
        any setting changes since the first access.

        Type:
            str
        """
        return sha256(
            json.dumps(
                {
                    field.name: getattr(self, field.name)
                    for field in dataclass_fields(self)
                },
                sort_keys=True)
            .encode('utf-8')
        ).hexdigest()



"""Django app information for reviewboard.scmtools.

Version Added:
    5.0
"""

from django.apps import AppConfig


class SCMToolsAppConfig(AppConfig):
    """App configuration for reviewboard.scmtools.

    Version Added:
        5.0
    """

    name = 'reviewboard.scmtools'

    def ready(self):
        """Configure the app once it's ready.

        This will connect signal handlers needed for repository and SCMTool
        management.
        """
        from reviewboard.scmtools.signal_handlers import \
            connect_signal_handlers

        connect_signal_handlers()



"""Add an index for ReviewRequest.summary.

Version Added:
    2.0.9
"""

from django_evolution.mutations import ChangeField


MUTATIONS = [
    ChangeField('ReviewRequest', 'summary', initial=None, db_index=True),
]



"""DiffSet model definition."""

from __future__ import annotations

from typing import ClassVar

from django.core.exceptions import ValidationError
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext, gettext_lazy as _
from djblets.db.fields import JSONField, RelationCounterField

from reviewboard.diffviewer.filediff_creator import create_filediffs
from reviewboard.diffviewer.diffutils import get_total_line_counts
from reviewboard.diffviewer.managers import DiffSetManager
from reviewboard.scmtools.models import Repository


class DiffSet(models.Model):
    """A revisioned collection of FileDiffs."""

    _FINALIZED_COMMIT_SERIES_KEY = '__finalized_commit_series'

    name = models.CharField(_('name'), max_length=256)
    revision = models.IntegerField(_('revision'))
    timestamp = models.DateTimeField(_('timestamp'), default=timezone.now)
    basedir = models.CharField(_('base directory'), max_length=256,
                               blank=True, default='')
    history = models.ForeignKey('DiffSetHistory',
                                on_delete=models.CASCADE,
                                null=True,
                                related_name='diffsets',
                                verbose_name=_('diff set history'))
    repository = models.ForeignKey(Repository,
                                   on_delete=models.CASCADE,
                                   related_name='diffsets',
                                   verbose_name=_('repository'))
    diffcompat = models.IntegerField(
        _('differ compatibility version'),
        default=0,
        help_text=_('The diff generator compatibility version to use. '
                    'This can and should be ignored.'))

    base_commit_id = models.CharField(
        _('commit ID'), max_length=64, blank=True, null=True, db_index=True,
        help_text=_('The ID/revision this change is built upon.'))

    commit_count = RelationCounterField('commits')

    extra_data = JSONField(null=True)

    objects: ClassVar[DiffSetManager] = DiffSetManager()

    @property
    def is_commit_series_finalized(self):
        """Whether the commit series represented by this DiffSet is finalized.

        When a commit series is finalized, no more :py:class:`DiffCommits
        <reviewboard.diffviewer.models.diffcommit.DiffCommit>` can be added to
        it.
        """
        return (self.extra_data and
                self.extra_data.get(self._FINALIZED_COMMIT_SERIES_KEY, False))

    def finalize_commit_series(self, cumulative_diff, validation_info,
                               parent_diff=None, request=None, validate=True,
                               save=False):
        """Finalize the commit series represented by this DiffSet.

        Args:
            cumulative_diff (bytes):
                The cumulative diff of the entire commit series.

            validation_info (dict):
                The parsed validation information.

            parent_diff (bytes, optional):
                The parent diff of the cumulative diff, if any.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client, if any.

            validate (bool, optional):
                Whether or not the cumulative diff (and optional parent diff)
                should be validated, up to and including file existence checks.

            save (bool, optional):
                Whether to save the model after finalization. Defaults to
                ``False``.

                If ``True``, only the :py:attr:`extra_data` field will be
                updated.

                If ``False``, the caller must save this model.

        Returns:
            list of reviewboard.diffviewer.models.filediff.FileDiff:
            The list of created FileDiffs.

        Raises:
            django.core.exceptions.ValidationError:
                The commit series failed validation.
        """
        if validate:
            if self.is_commit_series_finalized:
                raise ValidationError(
                    gettext('This diff is already finalized.'),
                    code='invalid')

            if not self.files.exists():
                raise ValidationError(
                    gettext('Cannot finalize an empty commit series.'),
                    code='invalid')

            commits = {
                commit.commit_id: commit
                for commit in self.commits.all()
            }

            missing_commit_ids = set()

            for commit_id, info in validation_info.items():
                if (commit_id not in commits or
                    commits[commit_id].parent_id != info['parent_id']):
                    missing_commit_ids.add(commit_id)

            if missing_commit_ids:
                raise ValidationError(
                    gettext('The following commits are specified in '
                            'validation_info but do not exist: %s')
                    % ', '.join(missing_commit_ids),
                    code='validation_info')

            for commit_id, commit in commits.items():
                if (commit_id not in validation_info or
                    validation_info[commit_id]['parent_id'] !=
                        commit.parent_id):
                    missing_commit_ids.add(commit_id)

            if missing_commit_ids:
                raise ValidationError(
                    gettext('The following commits exist but are not '
                            'present in validation_info: %s')
                    % ', '.join(missing_commit_ids),
                    code='validation_info')

        filediffs = create_filediffs(
            get_file_exists=self.repository.get_file_exists,
            diff_file_contents=cumulative_diff,
            parent_diff_file_contents=parent_diff,
            repository=self.repository,
            request=request,
            basedir=self.basedir,
            base_commit_id=self.base_commit_id,
            diffset=self,
            check_existence=validate)

        if self.extra_data is None:
            self.extra_data = {}

        self.extra_data[self._FINALIZED_COMMIT_SERIES_KEY] = True

        if save:
            self.save(update_fields=('extra_data',))

        return filediffs

    def get_total_line_counts(self):
        """Return the total line counts of all child FileDiffs.

        Returns:
            dict:
            A dictionary with the following keys:

            * ``raw_insert_count``
            * ``raw_delete_count``
            * ``insert_count``
            * ``delete_count``
            * ``replace_count``
            * ``equal_count``
            * ``total_line_count``

            Each entry maps to the sum of that line count type for all child
            :py:class:`FileDiffs
            <reviewboard.diffviewer.models.filediff.FileDiff>`.
        """
        return get_total_line_counts(self.files.all())

    @property
    def per_commit_files(self):
        """The files limited to per-commit diffs.

        This will cache the results for future lookups. If the set of all files
        has already been fetched with :py:meth:`~django.db.models.query.
        QuerySet.prefetch_related`, no queries will be performed.
        """
        if not hasattr(self, '_per_commit_files'):
            # This is a giant hack because Django 1.6.x does not support
            # Prefetch() statements. In Django 1.8+ we can replace any use of:
            #
            #     # Django == 1.6
            #     ds = DiffSet.objects.prefetch_related('files')
            #     for d in ds:
            #         # Do something with d.per_commit_files
            #
            # with:
            #
            #     # Django >= 1.8
            #     ds = DiffSet.objects.prefetch_related(
            #         Prefetch('files',
            #                  queryset=File.objects.filter(
            #                      commit_id__isnull=False),
            #                  to_attr='per_commit_files')
            #     for d in ds:
            #         # Do something with d.per_commit_files
            if (hasattr(self, '_prefetched_objects_cache') and
                'files' in self._prefetched_objects_cache):
                self._per_commit_files = sorted(
                    (
                        f
                        for f in self.files.all()
                        if f.commit_id is not None
                    ),
                    key=lambda f: f.pk)
            else:
                self._per_commit_files = list(
                    self.files
                    .filter(commit_id__isnull=False)
                    .order_by('pk')
                )

        return self._per_commit_files

    @property
    def cumulative_files(self):
        """The files limited to the cumulative diff.

        This will cache the results for future lookups. If the set of all files
        has been already been fetched with :py:meth:`~django.db.models.query.
        QuerySet.prefetch_related`, no queries will be incurred.
        """
        # See per_commit_files for why we are doing this hack.
        if not hasattr(self, '_cumulative_files'):
            if (hasattr(self, '_prefetched_objects_cache') and
                'files' in self._prefetched_objects_cache):
                self._cumulative_files = sorted(
                    (
                        f
                        for f in self.files.all()
                        if f.commit_id is None
                    ),
                    key=lambda f: f.pk)
            else:
                self._cumulative_files = list(
                    self.files
                    .filter(commit_id__isnull=True)
                    .order_by('pk')
                )

        return self._cumulative_files

    def update_revision_from_history(self, diffset_history):
        """Update the revision of this diffset based on a diffset history.

        This will determine the appropriate revision to use for the diffset,
        based on how many other diffsets there are in the history. If there
        aren't any, the revision will be set to 1.

        Args:
            diffset_history (reviewboard.diffviewer.models.diffset_history.
                             DiffSetHistory):
                The diffset history used to compute the new revision.

        Raises:
            ValueError:
                The revision already has a valid value set, and cannot be
                updated.
        """
        if self.revision not in (0, None):
            raise ValueError('The diffset already has a valid revision set.')

        # Default this to revision 1. We'll use this if the DiffSetHistory
        # isn't saved yet (which may happen when creating a new review request)
        # or if there aren't yet any diffsets.
        self.revision = 1

        if diffset_history.pk:
            try:
                latest_diffset = \
                    diffset_history.diffsets.only('revision').latest()
                self.revision = latest_diffset.revision + 1
            except DiffSet.DoesNotExist:
                # Stay at revision 1.
                pass

    def save(self, **kwargs):
        """Save this diffset.

        This will set an initial revision of 1 if this is the first diffset
        in the history, and will set it to on more than the most recent
        diffset otherwise.

        Args:
            **kwargs (dict):
                Extra arguments for the save call.
        """
        if self.history is not None:
            if self.revision == 0:
                self.update_revision_from_history(self.history)

            self.history.last_diff_updated = self.timestamp
            self.history.save()

        super(DiffSet, self).save(**kwargs)

    def __str__(self):
        """Return a human-readable representation of the DiffSet.

        Returns:
            unicode:
            A human-readable representation of the DiffSet.
        """
        return "[%s] %s r%s" % (self.id, self.name, self.revision)

    class Meta:
        app_label = 'diffviewer'
        db_table = 'diffviewer_diffset'
        get_latest_by = 'revision'
        ordering = ['revision', 'timestamp']
        verbose_name = _('Diff Set')
        verbose_name_plural = _('Diff Sets')



"""Actions framework for Review Board.

Version Added:
    6.0
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from djblets.registries.importer import lazy_import_registry

from reviewboard.actions.base import (ActionAttachmentPoint,
                                      ActionPlacement,
                                      AttachmentPoint,
                                      BaseAction,
                                      BaseGroupAction,
                                      BaseMenuAction)

if TYPE_CHECKING:
    from reviewboard.actions.registry import (
        ActionAttachmentPointsRegistry,
        ActionsRegistry,
    )


#: The actions registry.
actions_registry: ActionsRegistry = lazy_import_registry(
    'reviewboard.actions.registry',
    'ActionsRegistry')  # type: ignore


#: The action attachment points registry.
#:
#: Version Added:
#:     8.0
action_attachment_points_registry: ActionAttachmentPointsRegistry = \
    lazy_import_registry('reviewboard.actions.registry',
                         'ActionAttachmentPointsRegistry')  # type: ignore


__all__ = [
    'ActionAttachmentPoint',
    'ActionPlacement',
    'AttachmentPoint',
    'BaseAction',
    'BaseGroupAction',
    'BaseMenuAction',
    'action_attachment_points_registry',
    'actions_registry',
]

__autodoc_excludes__ = [
    'ActionAttachmentPoint',
    'ActionPlacement',
    'AttachmentPoint',
    'BaseAction',
    'BaseGroupAction',
    'BaseMenuAction',
]



"""URLs for the accounts app."""

from django.conf import settings
from django.urls import include, path, re_path
from django.contrib.auth import views as auth_views
from djblets.urls.resolvers import DynamicURLResolver

from reviewboard.accounts.forms.auth import AuthenticationForm
from reviewboard.accounts.views import LoginView
from reviewboard.accounts import views as accounts_views


sso_dynamic_urls = DynamicURLResolver()


urlpatterns = [
    path('login/',
         LoginView.as_view(
             template_name='accounts/login.html',
             authentication_form=AuthenticationForm),
         name='login'),
    path('logout/',
         auth_views.LogoutView.as_view(
             template_name='accounts/logged_out.html',
         ),
         name='logout'),
    path('preferences/',
         accounts_views.MyAccountView.as_view(),
         name='user-preferences'),
    path('register/',
         accounts_views.account_register,
         kwargs={
             'next_url': 'dashboard',
         },
         name='register'),
    path('recover/',
         auth_views.PasswordResetView.as_view(
             template_name='accounts/password_reset.html',
             email_template_name='accounts/password_reset_email.txt',
             extra_email_context={
                 'PRODUCT_NAME': settings.PRODUCT_NAME,
             }),
         name='recover'),
    path('recover/done/',
         auth_views.PasswordResetDoneView.as_view(
             template_name='accounts/password_reset_done.html'),
         name='password_reset_done'),
    path('reset/<uidb64>/<token>/',
         auth_views.PasswordResetConfirmView.as_view(
             template_name='accounts/password_reset_confirm.html'),
         name='password_reset_confirm'),
    path('reset/done/',
         auth_views.PasswordResetCompleteView.as_view(
             template_name='accounts/password_reset_complete.html'),
         name='password_reset_complete'),
    re_path(r'^preferences/oauth2-application/(?:(?P<app_id>[0-9]+)/)?$',
            accounts_views.edit_oauth_app,
            name='edit-oauth-app'),
    path('sso/', include(([sso_dynamic_urls], 'accounts'), namespace='sso')),
    path('client-login/',
         accounts_views.ClientLoginView.as_view(),
         name='client-login'),
    path('client-login/confirm',
         accounts_views.ClientLoginConfirmationView.as_view(),
         name='client-login-confirm'),
]


if settings.DEBUG and not settings.PRODUCTION:
    urlpatterns += [
        re_path(r'^preferences/preview-email/password-changed/'
                r'(?P<message_format>text|html)/$',
                accounts_views.preview_password_changed_email,
                name='preview-password-change-email'),
        path('preferences/preview-email/password-changed/',
             accounts_views.preview_password_changed_email,
             name='preview-password-change-email'),
    ]



import logging
import re

from django import template
from django.template.loader import render_to_string
from django.utils.html import escape, format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

from reviewboard.codesafety import code_safety_checker_registry
from reviewboard.diffviewer.chunk_generator import DiffChunkGenerator


logger = logging.getLogger(__name__)
register = template.Library()


@register.filter
def highlightregion(value, regions):
    """Highlights the specified regions of text.

    This is used to insert ``<span class="hl">...</span>`` tags in the
    text as specified by the ``regions`` variable.
    """
    if not regions:
        return value

    html = []

    # We need to insert span tags into a string already consisting
    # of span tags. We have a list of ranges that our span tags should
    # go into, but those ranges are in the markup-less string.
    #
    # We go through the string and keep track of the location in the
    # markup and in the markup-less string. We make sure to insert our
    # span tag any time that we're within the current region, so long
    # as we haven't already created one. We close the span tag whenever
    # we're done with the region or when we're about to enter a tag in
    # the markup string.
    #
    # This code makes the assumption that the list of regions is sorted.
    # This is safe to assume in practice, but if we ever at some point
    # had reason to doubt it, we could always sort the regions up-front.
    in_hl = False
    i = j = r = 0
    region_start, region_end = regions[r]

    while i < len(value):
        c = value[i]

        if c == '<':
            if in_hl:
                html.append('</span>')
                in_hl = False

            k = value.find('>', i)
            assert k != -1

            html.append(value[i:k + 1])
            i = k
        else:
            if not in_hl and region_start <= j < region_end:
                html.append('<span class="hl">')
                in_hl = True

            if c == '&':
                k = value.find(';', i)
                assert k != -1

                html.append(value[i:k + 1])
                i = k
                j += 1
            else:
                j += 1
                html.append(c)

        if j == region_end:
            if in_hl:
                html.append('</span>')
                in_hl = False

            r += 1

            if r == len(regions):
                break

            region_start, region_end = regions[r]

        i += 1

    if i + 1 < len(value):
        html.append(value[i + 1:])

    return mark_safe(''.join(html))


extraWhitespace = re.compile(r'(\s+(</span>)?$| +\t)')


@register.filter
def showextrawhitespace(value):
    """
    Marks up any extra whitespace in the specified text.

    Any trailing whitespace or tabs following one or more spaces are
    marked up by inserted ``<span class="ew">...</span>`` tags.
    """
    value = extraWhitespace.sub(r'<span class="ew">\1</span>', value)
    return mark_safe(value.replace('\t', '<span class="tb">\t</span>'))


def _diff_expand_link(context, expandable, text, tooltip,
                      expand_pos, image_class):
    """Utility function to render a diff expansion link.

    This is used internally by other template tags to provide a diff
    expansion link. It assumes nothing about the content and serves only
    to render the data from a template.
    """
    return render_to_string(
        template_name='diffviewer/expand_link.html',
        context={
            'tooltip': tooltip,
            'text': text,
            'chunk': context['chunk'],
            'file': context['file'],
            'expand_pos': expand_pos,
            'image_class': image_class,
            'expandable': expandable,
        })


@register.simple_tag(takes_context=True)
def diff_expand_link(context, expanding, tooltip,
                     expand_pos_1=None, expand_pos_2=None, text=None):
    """Renders a diff expansion link.

    This link will expand the diff entirely, or incrementally in one
    or more directions.

    'expanding' is expected to be one of 'all', 'above', or 'below'.
    """
    if expanding == 'all':
        image_class = 'rb-icon-diff-expand-all'
        expand_pos = None
    else:
        lines_of_context = context['lines_of_context']
        expand_pos = (lines_of_context[0] + expand_pos_1,
                      lines_of_context[1] + expand_pos_2)
        image_class = 'rb-icon-diff-expand-%s' % expanding

    return _diff_expand_link(context, True, text, tooltip, expand_pos,
                             image_class)


@register.simple_tag(takes_context=True)
def diff_chunk_header(context, header):
    """Renders a diff header as HTML.

    This diff header may be expandable, depending on whether or not the
    function/class referenced in the header is contained within the collapsed
    region.
    """
    lines_of_context = context['lines_of_context']
    chunk = context['chunk']

    line = chunk['lines'][0]

    if line[1] != '' and header['line'] >= line[1]:
        expand_offset = line[1] + chunk['numlines'] - header['line']
        expandable = True
    else:
        expand_offset = 0
        expandable = False

    return _diff_expand_link(context, expandable,
                             '<code>%s</code>' % escape(header['text']),
                             _('Expand to header'),
                             (lines_of_context[0],
                              expand_offset + lines_of_context[1]),
                             'rb-icon-diff-expand-header')


@register.simple_tag
def diff_lines(index, chunk, standalone, line_fmt, anchor_fmt='',
               begin_collapse_fmt='', end_collapse_fmt='', moved_fmt='',
               line_warnings_fmt=''):
    """Renders the lines of a diff.

    This will render each line in the diff viewer. The function expects
    some basic data on what will be rendered, as well as printf-formatted
    templates for the contents.

    Python ``%``-formatted templates are used instead of standard Django
    templates because they're much faster to render, which makes a huge
    difference when rendering thousands of lines or more.

    Version Changed:
        5.0:
        Added ``line_warnings_fmt``.

    Args:
        index (str):
            The index of the chunk.

        chunk (dict):
            The chunk information.

        standalone (bool):
            Whether to generate just the specific chunk index. This is used
            for expanding/collapsing.

        line_fmt (str):
            The ``%``-formatted template for a side-by-side line.

        anchor_fmt (str, optional):
            The Python ``%``-formatted template for an anchor to a line.

        begin_collapse_fmt (str, optional):
            The Python ``%``-formatted template for the beginning of a
            collapsed section.

        end_collapse_fmt (str, optional):
            The Python ``%``-formatted template for the end of a collapsed
            section.

        moved_fmt (str, optional):
            The Python ``%``-formatted template for a move flag.

        line_warnings_fmt (str, optional):
            The Python ``%``-formatted template for a line warnings indicator.

            Version Added:
                5.0

    Returns:
        django.utils.safestring.SafeString:
        The rendered HTML.
    """
    STYLED_MAX_LINE_LEN = DiffChunkGenerator.STYLED_MAX_LINE_LEN

    lines = chunk['lines']
    num_lines = len(lines)
    chunk_index = chunk['index']
    change = chunk['change']
    is_equal = False
    is_replace = False
    is_insert = False
    is_delete = False

    if change == 'equal':
        is_equal = True
    elif change == 'replace':
        is_replace = True
    elif change == 'insert':
        is_insert = True
    elif change == 'delete':
        is_delete = True

    result = []
    last_line_num = num_lines - 1

    for i, line in enumerate(lines):
        row_classes = []
        header_1_classes = []
        header_2_classes = []
        cell_1_classes = ['l']
        cell_2_classes = ['r']
        row_class_attr = ''
        header_1_class_attr = ''
        header_2_class_attr = ''
        cell_1_class_attr = ''
        cell_2_class_attr = ''
        line_html_sides = [line[2], line[5]]
        linenum1 = line[1]
        linenum2 = line[4]
        show_collapse = False
        anchor = None
        warning_labels = set()

        try:
            line_meta = line[8]
        except IndexError:
            line_meta = {}

        code_safety_results = line_meta.get('code_safety', [])

        if i == 0:
            row_classes.append('first')

        if i == last_line_num:
            row_classes.append('last')

        if not is_equal:
            if i == 0:
                anchor = '%s.%s' % (index, chunk_index)

            if line[7]:
                row_classes.append('whitespace-line')
        else:
            show_collapse = (i == 0 and standalone)

        # Conditionally update any content on either side of the displayed
        # line. Only do this for lines that would contain changes (inserted,
        # deleted, replaced, or equal lines).
        for side_i, process, line_range in ((0, not is_insert, line[3]),
                                            (1, not is_delete, line[6])):
            if not process:
                continue

            line_html = line_html_sides[side_i]

            if line_html and len(line_html) <= STYLED_MAX_LINE_LEN:
                # NOTE: It's important that highlighting occurs before any
                #       code that may modify the length of any text content
                #       within tags, or the highlighting regions will be
                #       incorrect.
                if is_replace and line_range:
                    line_html = highlightregion(line_html, line_range)

                line_html = showextrawhitespace(line_html)

                for checker_id, checker_results in code_safety_results:
                    checker_result_ids = sorted(
                        checker_results.get('warnings', set()) |
                        checker_results.get('errors', set()))
                    assert checker_result_ids

                    checker = code_safety_checker_registry.get_checker(
                        checker_id)

                    if checker is None:
                        # This is very unlikely, but could happen if a checker
                        # is provided by an extension (or an older version of
                        # Review Board) and is not available on the machine
                        # rendering cached diff chunk results.
                        #
                        # Fall back to providing the raw result codes.
                        warning_labels.update(checker_result_ids)
                    else:
                        line_html = checker.update_line_html(
                            line_html=line_html,
                            result_ids=checker_result_ids)
                        warning_labels.update(checker.get_result_labels(
                            checker_result_ids))

                line_html_sides[side_i] = line_html

        # Check for any move information. If found, prepare CSS classes and
        # HTML to show on the line.
        moved_from = {}
        moved_to = {}
        is_moved_row = False
        is_first_moved_row = False

        if line_meta:
            if 'from' in line_meta:
                moved_from_linenum, moved_from_first = line_meta['from']
                is_moved_row = True

                header_2_classes.append('moved-from')
                cell_2_classes.append('moved-from')

                if moved_from_first:
                    # This is the start of a new move range.
                    is_first_moved_row = True
                    header_2_classes.append('moved-from-start')
                    cell_2_classes.append('moved-from-start')
                    moved_from = {
                        'class': 'moved-flag',
                        'line': mark_safe('moved-from-%s'
                                          % moved_from_linenum),
                        'target': mark_safe('moved-to-%s' % linenum2),
                        'text': _('Moved from line %s') % moved_from_linenum,
                    }

            if 'to' in line_meta:
                moved_to_linenum, moved_to_first = line_meta['to']
                is_moved_row = True

                header_1_classes.append('moved-to')
                cell_1_classes.append('moved-to')

                if moved_to_first:
                    # This is the start of a new move range.
                    is_first_moved_row = True
                    header_1_classes.append('moved-to-start')
                    cell_1_classes.append('moved-to-start')
                    moved_to = {
                        'class': 'moved-flag',
                        'line': mark_safe('moved-to-%s' % moved_to_linenum),
                        'target': mark_safe('moved-from-%s' % linenum1),
                        'text': _('Moved to line %s') % moved_to_linenum,
                    }

        # Build all the attributes that will be used for the elements.
        if is_moved_row:
            row_classes.append('moved-row')

        if is_first_moved_row:
            row_classes.append('moved-row-start')

        if row_classes:
            row_class_attr = ' class="%s"' % ' '.join(row_classes)

        if cell_1_classes:
            cell_1_class_attr = ' class="%s"' % ' '.join(cell_1_classes)

        if cell_2_classes:
            cell_2_class_attr = ' class="%s"' % ' '.join(cell_2_classes)

        if header_1_classes:
            header_1_class_attr = ' class="%s"' % ' '.join(header_1_classes)

        if header_2_classes:
            header_2_class_attr = ' class="%s"' % ' '.join(header_2_classes)

        # Build all the HTML used for the line.
        anchor_html = ''
        begin_collapse_html = ''
        end_collapse_html = ''
        moved_from_html = ''
        moved_to_html = ''
        warnings_html = ''

        context = {
            'chunk_index': chunk_index,
            'row_class_attr': row_class_attr,
            'header_1_class_attr': header_1_class_attr,
            'header_2_class_attr': header_2_class_attr,
            'cell_1_class_attr': cell_1_class_attr,
            'cell_2_class_attr': cell_2_class_attr,
            'linenum_row': line[0],
            'linenum1': linenum1,
            'linenum2': linenum2,
            'line1': line_html_sides[0],
            'line2': line_html_sides[1],
            'moved_from': moved_from,
            'moved_to': moved_to,
        }

        if anchor:
            anchor_html = anchor_fmt % {
                'anchor': anchor,
            }

        if show_collapse:
            begin_collapse_html = begin_collapse_fmt % context
            end_collapse_html = end_collapse_fmt % context

        if moved_from:
            moved_from_html = moved_fmt % moved_from

        if moved_to:
            moved_to_html = moved_fmt % moved_to

        if warning_labels:
            warnings_html = line_warnings_fmt % {
                'warning_labels': ', '.join(
                    escape(_label)
                    for _label in warning_labels
                ),
            }

        context.update({
            'anchor_html': anchor_html,
            'begin_collapse_html': begin_collapse_html,
            'end_collapse_html': end_collapse_html,
            'moved_from_html': moved_from_html,
            'moved_to_html': moved_to_html,
            'warnings_html': warnings_html,
        })

        result.append(line_fmt % context)

    return mark_safe(''.join(result))


@register.simple_tag
def diff_code_safety_file_alert(checker_id, checker_results):
    """Render a code safety alert at the top of a file.

    Version Added:
        5.0

    Args:
        checker_id (str):
            The ID of the code safety checker.

        checker_results (dict):
            The list of code safety checker results. This may contain
            ``errors`` and ``warnings`` keys spanning any errors/warnings
            found throughout the file by this checker.

    Returns:
        django.utils.safestring.SafeString:
        The rendered HTML.
    """
    checker = code_safety_checker_registry.get_checker(checker_id)

    if checker is None:
        # This is very unlikely, but could happen if a checker is provided
        # by an extension (or an older version of Review Board) and is not
        # available on the machine rendering cached diff chunk results.
        logger.error(
            'Error locating code safety checker "%s" when attempting to '
            'render code safety file alert.',
            checker_id)

        return format_html(
            _('A code safety checker ({checker_id}) previously found '
              'problems, but could not be loaded while rendering this diff. '
              'It returned the following results: {results}'),
            checker_id=checker_id,
            results=checker_results)

    return checker.render_file_alert_html(
        error_ids=checker_results.get('errors'),
        warning_ids=checker_results.get('warnings'))



from django.contrib.auth.models import User
from djblets.features.testing import override_feature_check
from djblets.webapi.errors import INVALID_FORM_DATA, PERMISSION_DENIED
from djblets.webapi.testing.decorators import webapi_test_template
from kgb import SpyAgency

from reviewboard.diffviewer.features import dvcs_feature
from reviewboard.diffviewer.models import FileDiff
from reviewboard.reviews.models import Comment
from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import (
    review_diff_comment_item_mimetype,
    review_diff_comment_list_mimetype)
from reviewboard.webapi.tests.mixins import (
    BasicTestsMetaclass,
    ReviewRequestChildItemMixin,
    ReviewRequestChildListMixin)
from reviewboard.webapi.tests.mixins_comment import (
    CommentItemMixin,
    CommentListMixin)
from reviewboard.webapi.tests.urls import (
    get_review_diff_comment_item_url,
    get_review_diff_comment_list_url)


class BaseResourceTestCase(BaseWebAPITestCase):
    def _common_post_interdiff_comments(self, comment_text):
        review_request, filediff = self._create_diff_review_request()
        diffset = filediff.diffset

        # Post the second diff.
        interdiffset = self.create_diffset(review_request)
        interfilediff = self.create_filediff(diffset)

        review = self.create_review(review_request, user=self.user)
        comment = self.create_diff_comment(review, filediff, interfilediff,
                                           text=comment_text)

        return comment, review_request, review, interdiffset.revision

    def _create_diff_review_with_issue(self, publish=False, comment_text=None,
                                       expected_status=201):
        """Sets up a review for a diff that includes a comment with an issue.

        If `publish` is True, the review is published. The review request is
        always published.

        Returns the response from posting the comment, the review object, and
        the review request object.
        """
        if not comment_text:
            comment_text = 'Test diff comment with an opened issue'

        review_request, filediff = self._create_diff_review_request()
        review = self.create_review(review_request, user=self.user,
                                    publish=publish)
        comment = self.create_diff_comment(review, filediff, text=comment_text,
                                           issue_opened=True)

        return comment, review, review_request

    def _create_diff_review_request(self, with_local_site=False,
                                    with_history=False, repository=None):
        """Create a review request with a diff for testing.

        Version Changed:
            5.0:
            Added the ``repository`` argument.

        Args:
            with_local_site (bool, optional):
                Whether to create the review request on a Local Site.

            with_history (bool, optional):
                Whether to create the review request with multi-commit
                history enabled.

            repository (reviewboard.scmtools.models.Repository, optional):
                An explicit repository to attach to the review request.

                If not provided, one will be created.

                Version Added:
                    5.0

        Returns:
            tuple:
            A 2-tuple of:

            1. The review request instance.
            2. The FileDiff instance.
        """
        review_request = self.create_review_request(
            create_repository=repository is None,
            repository=repository,
            submitter=self.user,
            with_local_site=with_local_site,
            create_with_history=with_history,
            publish=True)
        diffset = self.create_diffset(review_request)

        if with_history:
            commit = self.create_diffcommit(diffset=diffset)
        else:
            commit = None

        filediff = self.create_filediff(diffset=diffset, commit=commit)

        return review_request, filediff

    def _create_diff_review(self):
        review_request, filediff = self._create_diff_review_request()

        review = self.create_review(review_request, publish=True)
        self.create_diff_comment(review, filediff)

        return review


def _compare_item(self, item_rsp, comment):
    """Compare the API response with the object that was serialized.

    Args:
        item_rsp (dict):
            The serialized comment.

        comment (reviewboard.reviews.models.diff_comment.Comment):
            The comment that was serialized.

    Raises:
        AssertionError:
            The API response was not equivalent to the object.
    """
    self.assertEqual(item_rsp['id'], comment.pk)
    self.assertEqual(item_rsp['text'], comment.text)
    self.assertEqual(item_rsp['issue_opened'], comment.issue_opened)
    self.assertEqual(item_rsp['first_line'], comment.first_line)
    self.assertEqual(item_rsp['num_lines'], comment.num_lines)

    self.assertEqual(item_rsp['extra_data'],
                     self.resource._strip_private_data(comment.extra_data))

    if comment.rich_text:
        self.assertEqual(item_rsp['text_type'], 'markdown')
    else:
        self.assertEqual(item_rsp['text_type'], 'plain')


class ResourceListTests(SpyAgency, CommentListMixin,
                        ReviewRequestChildListMixin, BaseResourceTestCase,
                        metaclass=BasicTestsMetaclass):
    """Testing the ReviewDiffCommentResource list APIs."""
    fixtures = ['test_users', 'test_scmtools']
    sample_api_url = 'review-requests/<id>/reviews/<id>/diff-comments/'
    resource = resources.review_diff_comment

    compare_item = _compare_item

    def setup_review_request_child_test(self, review_request):
        if not review_request.repository_id:
            # The group tests don't create a repository by default.
            review_request.repository = self.create_repository()
            review_request.save()

        diffset = self.create_diffset(review_request)
        self.create_filediff(diffset)
        review = self.create_review(review_request, publish=True)

        return (get_review_diff_comment_list_url(review),
                review_diff_comment_list_mimetype)

    def setup_http_not_allowed_item_test(self, user):
        review_request = self.create_review_request(
            create_repository=True,
            submitter=user,
            publish=True)
        review = self.create_review(review_request, publish=True)

        return get_review_diff_comment_list_url(review)
    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name,
                             populate_items):
        review_request = self.create_review_request(
            create_repository=True,
            with_local_site=with_local_site,
            submitter=user,
            publish=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)
        review = self.create_review(review_request, publish=True)

        if populate_items:
            items = [self.create_diff_comment(review, filediff)]
        else:
            items = []

        return (get_review_diff_comment_list_url(review, local_site_name),
                review_diff_comment_list_mimetype,
                items)

    def test_get_with_counts_only(self):
        """Testing the
        GET review-requests/<id>/reviews/<id>/diff-comments/?counts-only=1 API
        """
        review = self._create_diff_review()

        rsp = self.api_get(get_review_diff_comment_list_url(review), {
            'counts-only': 1,
        }, expected_mimetype=review_diff_comment_list_mimetype)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['count'], review.comments.count())

    def test_get_with_interdiff(self):
        """Testing the GET review-requests/<id>/reviews/<id>/diff-comments/ API
        with interdiff
        """
        comment_text = "Test diff comment"

        comment, review_request, review, interdiff_revision = \
            self._common_post_interdiff_comments(comment_text)

        rsp = self.api_get(get_review_diff_comment_list_url(review), {
            'interdiff-revision': interdiff_revision,
        }, expected_mimetype=review_diff_comment_list_mimetype)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('diff_comments', rsp)
        self.assertEqual(len(rsp['diff_comments']), 1)
        self.assertEqual(rsp['diff_comments'][0]['text'], comment_text)

    #
    # HTTP POST tests
    #

    def setup_basic_post_test(self, user, with_local_site, local_site_name,
                              post_valid_data):
        review_request = self.create_review_request(
            create_repository=True,
            with_local_site=with_local_site,
            submitter=user,
            publish=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)
        review = self.create_review(review_request, user=user)

        return (
            get_review_diff_comment_list_url(review, local_site_name),
            review_diff_comment_item_mimetype,
            {
                'filediff_id': filediff.pk,
                'text': 'My new text',
                'first_line': 1,
                'num_lines': 2,
            },
            [review])

    def check_post_result(self, user, rsp, review):
        comment_rsp = rsp['diff_comment']
        self.assertEqual(comment_rsp['text'], 'My new text')
        self.assertEqual(comment_rsp['text_type'], 'plain')

        comment = Comment.objects.get(pk=comment_rsp['id'])
        self.compare_item(comment_rsp, comment)

    def test_post_with_issue(self):
        """Testing the
        POST review-requests/<id>/reviews/<id>/diff-comments/ API
        with an issue
        """
        diff_comment_text = 'Test diff comment with an opened issue'

        review_request, filediff = self._create_diff_review_request()
        review = self.create_review(review_request, user=self.user)
        rsp = self.api_post(
            get_review_diff_comment_list_url(review),
            {
                'filediff_id': filediff.pk,
                'issue_opened': True,
                'first_line': 1,
                'num_lines': 5,
                'text': diff_comment_text,
            },
            expected_mimetype=review_diff_comment_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('diff_comment', rsp)
        self.assertEqual(rsp['diff_comment']['text'], diff_comment_text)
        self.assertTrue(rsp['diff_comment']['issue_opened'])

    def test_post_with_interdiff(self):
        """Testing the
        POST review-requests/<id>/reviews/<id>/diff-comments/ API
        with interdiff
        """
        comment_text = "Test diff comment"

        review_request, filediff = self._create_diff_review_request()

        # Post the second diff.
        interdiffset = self.create_diffset(review_request)
        interfilediff = self.create_filediff(interdiffset)

        review = self.create_review(review_request, user=self.user)

        rsp = self.api_post(
            get_review_diff_comment_list_url(review),
            {
                'filediff_id': filediff.pk,
                'interfilediff_id': interfilediff.pk,
                'issue_opened': True,
                'first_line': 1,
                'num_lines': 5,
                'text': comment_text,
            },
            expected_mimetype=review_diff_comment_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('diff_comment', rsp)
        self.assertEqual(rsp['diff_comment']['text'], comment_text)

        comment = Comment.objects.get(pk=rsp['diff_comment']['id'])
        self.assertEqual(comment.filediff_id, filediff.pk)
        self.assertEqual(comment.interfilediff_id, interfilediff.pk)

    @webapi_test_template
    def test_post_with_interfilediff_same_filediff(self):
        """Testing the POST <URL> API with interfilediff_id == filediff_id"""
        review_request, filediff = self._create_diff_review_request()
        review = self.create_review(review_request, user=self.user)

        rsp = self.api_post(
            get_review_diff_comment_list_url(review),
            {
                'filediff_id': filediff.pk,
                'interfilediff_id': filediff.pk,
                'issue_opened': True,
                'first_line': 1,
                'num_lines': 5,
                'text': 'foo',
            },
            expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
        self.assertEqual(rsp['fields'], {
            'interfilediff_id': ['This cannot be the same as filediff_id.'],
        })

    @webapi_test_template
    def test_post_with_interfilediff_outside_diffset_history(self):
        """Testing the POST <URL> API with interfilediff_id corresponding to a
        FileDiff outside the current DiffSetHistory
        """
        repository = self.create_repository()

        review_request, filediff = self._create_diff_review_request(
            repository=repository)
        review = self.create_review(review_request, user=self.user)

        other_filediff = self._create_diff_review_request(
            repository=repository)[1]

        rsp = self.api_post(
            get_review_diff_comment_list_url(review),
            {
                'filediff_id': filediff.pk,
                'interfilediff_id': other_filediff.pk,
                'issue_opened': True,
                'first_line': 1,
                'num_lines': 5,
                'text': 'foo',
            },
            expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
        self.assertEqual(rsp['fields'], {
            'interfilediff_id': ['This is not a valid interfilediff ID.'],
        })

    @webapi_test_template
    def test_post_with_base_filediff_dvcs_enabled_with_history(self):
        """Testing the POST <URL> API with base_filediff_id with DVCS enabled
        on a review request created with commit history
        """
        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                create_with_history=True,
                publish=True)

            diffset = self.create_diffset(review_request)

            commits = [
                self.create_diffcommit(diffset=diffset, commit_id='r1',
                                       parent_id='r0'),
                self.create_diffcommit(diffset=diffset, commit_id='r2',
                                       parent_id='r1'),
            ]

            filediffs = [
                self.create_filediff(diffset=diffset, commit=commits[0],
                                     source_file='/foo', source_revision='1',
                                     dest_file='/foo', dest_detail='2'),
                self.create_filediff(diffset=diffset, commit=commits[1],
                                     source_file='/foo', source_revision='2',
                                     dest_file='/foo', dest_detail='3'),
            ]
            review = self.create_review(review_request, user=self.user)

            rsp = self.api_post(
                get_review_diff_comment_list_url(review),
                {
                    'filediff_id': filediffs[1].pk,
                    'base_filediff_id': filediffs[0].pk,
                    'issue_opened': True,
                    'first_line': 1,
                    'num_lines': 5,
                    'text': 'foo',
                },
                expected_mimetype=review_diff_comment_item_mimetype)

            self.assertEqual(rsp['stat'], 'ok')
            self.assertIn('diff_comment', rsp)

            item_rsp = rsp['diff_comment']
            comment = Comment.objects.get(pk=item_rsp['id'])

            self.compare_item(item_rsp, comment)

    @webapi_test_template
    def test_post_with_base_filediff_dvcs_disabled(self):
        """Testing the POST <URL> API with base_filediff_id when DVCS feature
        disabled
        """
        with override_feature_check(dvcs_feature.feature_id, enabled=False):
            review_request, filediff = self._create_diff_review_request()
            review = self.create_review(review_request, user=self.user)

            rsp = self.api_post(
                get_review_diff_comment_list_url(review),
                {
                    'filediff_id': filediff.pk,
                    'base_filediff_id': filediff.pk,
                    'issue_opened': True,
                    'first_line': 1,
                    'num_lines': 5,
                    'text': 'foo',
                },
                expected_mimetype=review_diff_comment_item_mimetype)

            self.assertEqual(rsp['stat'], 'ok')
            self.assertIn('diff_comment', rsp)

            item_rsp = rsp['diff_comment']
            comment = Comment.objects.get(pk=item_rsp['id'])

            self.compare_item(item_rsp, comment)

    @webapi_test_template
    def test_post_with_base_filediff_dvcs_enabled_no_history(self):
        """Testing the POST <URL> API with base_filediff_id when DVCS feature
        enabled and review request not created with commit history
        """
        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request, filediff = self._create_diff_review_request()
            review = self.create_review(review_request, user=self.user)

            rsp = self.api_post(
                get_review_diff_comment_list_url(review),
                {
                    'filediff_id': filediff.pk,
                    'base_filediff_id': filediff.pk,
                    'issue_opened': True,
                    'first_line': 1,
                    'num_lines': 5,
                    'text': 'foo',
                },
                expected_status=400)

            self.assertEqual(rsp['stat'], 'fail')
            self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
            self.assertEqual(
                rsp['fields']['base_filediff_id'],
                ['This field cannot be specified on review requests created '
                 'without history support.'])

    @webapi_test_template
    def test_post_with_base_filediff_dvcs_enabled_with_history_same_id(self):
        """Testing the POST <URL> API with base_filediff_id=filediff_id"""
        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request, filediff = self._create_diff_review_request(
                with_history=True)
            review = self.create_review(review_request, user=self.user)

            rsp = self.api_post(
                get_review_diff_comment_list_url(review),
                {
                    'filediff_id': filediff.pk,
                    'base_filediff_id': filediff.pk,
                    'issue_opened': True,
                    'first_line': 1,
                    'num_lines': 5,
                    'text': 'foo',
                },
                expected_status=400)

            self.assertEqual(rsp, {
                'stat': 'fail',
                'err': {
                    'code': INVALID_FORM_DATA.code,
                    'msg': INVALID_FORM_DATA.msg,
                    'type': INVALID_FORM_DATA.error_type,
                },
                'fields': {
                    'base_filediff_id': [
                        'This cannot be the same as filediff_id.',
                    ],
                },
            })

    @webapi_test_template
    def test_post_with_base_filediff_interdiff_dvcs_disabled(self):
        """Testing the POST <URL> API with base_filediff_id and interdiff_id
        when DVCS feature disabled
        """
        with override_feature_check(dvcs_feature.feature_id, enabled=False):
            review_request, filediff = self._create_diff_review_request()
            review = self.create_review(review_request, user=self.user)

            interdiffset = self.create_diffset(review_request)
            interfilediff = self.create_filediff(interdiffset)

            rsp = self.api_post(
                get_review_diff_comment_list_url(review),
                {
                    'filediff_id': filediff.pk,
                    'base_filediff_id': filediff.pk,
                    'interfilediff_id': interfilediff.pk,
                    'issue_opened': True,
                    'first_line': 1,
                    'num_lines': 5,
                    'text': 'foo',
                },
                expected_mimetype=review_diff_comment_item_mimetype)

            self.assertEqual(rsp['stat'], 'ok')
            self.assertIn('diff_comment', rsp)

            item_rsp = rsp['diff_comment']
            comment = Comment.objects.get(pk=item_rsp['id'])

            self.compare_item(item_rsp, comment)

    @webapi_test_template
    def test_post_with_base_filediff_interdiff_dvcs_enabled_with_history(self):
        """Testing the POST <URL> API with base_filediff_id and
        interfilediff_id when DVCS feature enabled
        """
        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request, filediff = self._create_diff_review_request(
                with_history=True)
            review = self.create_review(review_request, user=self.user)

            interdiffset = self.create_diffset(review_request)
            interfilediff = self.create_filediff(interdiffset)

            rsp = self.api_post(
                get_review_diff_comment_list_url(review),
                {
                    'filediff_id': filediff.pk,
                    'base_filediff_id': filediff.pk,
                    'interfilediff_id': interfilediff.pk,
                    'issue_opened': True,
                    'first_line': 1,
                    'num_lines': 5,
                    'text': 'foo',
                },
                expected_status=400)

            self.assertEqual(rsp['stat'], 'fail')
            self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
            self.assertEqual(rsp['fields'], {
                'base_filediff_id': [
                    'This field cannot be specified with interfilediff_id.',
                ],
                'interfilediff_id': [
                    'This field cannot be specified with base_filediff_id.',
                ],
            })

    @webapi_test_template
    def test_post_with_base_filediff_newer(self):
        """Testing the POST <URL> API with base_filediff_id newer than
        filediff_id
        """
        self.spy_on(FileDiff.get_ancestors,
                    owner=FileDiff)

        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                create_with_history=True,
                publish=True)

            diffset = self.create_diffset(review_request)

            commits = [
                self.create_diffcommit(diffset=diffset, commit_id='r1',
                                       parent_id='r0'),
                self.create_diffcommit(diffset=diffset, commit_id='r2',
                                       parent_id='r1'),
            ]

            filediffs = [
                self.create_filediff(diffset=diffset, commit=commit)
                for commit in commits
            ]
            review = self.create_review(review_request, user=self.user)

            rsp = self.api_post(
                get_review_diff_comment_list_url(review),
                {
                    'filediff_id': filediffs[0].pk,
                    'base_filediff_id': filediffs[1].pk,
                    'issue_opened': True,
                    'first_line': 1,
                    'num_lines': 5,
                    'text': 'foo',
                },
                expected_status=400)

            self.assertEqual(rsp['stat'], 'fail')
            self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
            self.assertEqual(rsp['fields'], {
                'base_filediff_id': [
                    'This is not a valid base filediff ID.',
                ],
            })

        self.assertFalse(FileDiff.get_ancestors.called)

    @webapi_test_template
    def test_post_with_base_filediff_same_commit(self):
        """Testing the POST <URL> API with base_filediff_id belonging to a
        different FileDiff in the same commit
        """
        self.spy_on(FileDiff.get_ancestors,
                    owner=FileDiff)

        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                create_with_history=True,
                publish=True)

            diffset = self.create_diffset(review_request)
            commit = self.create_diffcommit(diffset=diffset)
            filediffs = [
                self.create_filediff(diffset=diffset, commit=commit),
                self.create_filediff(diffset=diffset, commit=commit),
            ]
            review = self.create_review(review_request, user=self.user)

            rsp = self.api_post(
                get_review_diff_comment_list_url(review),
                {
                    'filediff_id': filediffs[0].pk,
                    'base_filediff_id': filediffs[1].pk,
                    'issue_opened': True,
                    'first_line': 1,
                    'num_lines': 5,
                    'text': 'foo',
                },
                expected_status=400)

            self.assertEqual(rsp['stat'], 'fail')
            self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
            self.assertEqual(rsp['fields'], {
                'base_filediff_id': [
                    'This is not a valid base filediff ID.',
                ]
            })

        self.assertFalse(FileDiff.get_ancestors.called)

    @webapi_test_template
    def test_post_with_base_filediff_not_exists(self):
        """Testing the POST <URL> API with base_filediff_id set to a
        non-existent ID
        """
        self.spy_on(FileDiff.get_ancestors,
                    owner=FileDiff)

        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request, filediff = self._create_diff_review_request(
                with_history=True)
            review = self.create_review(review_request, user=self.user)

            rsp = self.api_post(
                get_review_diff_comment_list_url(review),
                {
                    'filediff_id': filediff.pk,
                    'base_filediff_id': 12321,
                    'issue_opened': True,
                    'first_line': 1,
                    'num_lines': 5,
                    'text': 'foo',
                },
                expected_status=400)

            self.assertEqual(rsp['stat'], 'fail')
            self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
            self.assertEqual(rsp['fields'], {
                'base_filediff_id': [
                    'This is not a valid base filediff ID.',
                ]
            })

        self.assertFalse(FileDiff.get_ancestors.called)

    @webapi_test_template
    def test_post_with_base_filediff_outside_diffset(self):
        """Testing the POST <URL> API with base_filediff_id belonging to a
        different DiffSet
        """
        self.spy_on(FileDiff.get_ancestors,
                    owner=FileDiff)

        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                create_with_history=True,
                publish=True)

            diffsets = [
                self.create_diffset(review_request, revision=1),
                self.create_diffset(review_request, revision=2)
            ]

            commits = [
                self.create_diffcommit(diffset=diffsets[0], commit_id='r1',
                                       parent_id='r0'),
                self.create_diffcommit(diffset=diffsets[1], commit_id='r2',
                                       parent_id='r1'),
            ]

            filediffs = [
                self.create_filediff(diffset=diffset, commit=commit)
                for diffset, commit in zip(diffsets, commits)
            ]
            review = self.create_review(review_request, user=self.user)

            rsp = self.api_post(
                get_review_diff_comment_list_url(review),
                {
                    'filediff_id': filediffs[1].pk,
                    'base_filediff_id': filediffs[0].pk,
                    'issue_opened': True,
                    'first_line': 1,
                    'num_lines': 5,
                    'text': 'foo',
                },
                expected_status=400)

            self.assertEqual(rsp['stat'], 'fail')
            self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
            self.assertEqual(rsp['fields'], {
                'base_filediff_id': [
                    'This is not a valid base filediff ID.',
                ],
            })

        self.assertFalse(FileDiff.get_ancestors.called)

    @webapi_test_template
    def test_post_with_base_filediff_outside_history(self):
        """Testing the POST <URL> API with base_filediff_id not belonging to
        the FileDiff's set of ancestors
        """
        self.spy_on(FileDiff.get_ancestors,
                    owner=FileDiff)

        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                create_with_history=True,
                publish=True)

            diffset = self.create_diffset(review_request)

            commits = [
                self.create_diffcommit(diffset=diffset, commit_id='r1',
                                       parent_id='r0'),
                self.create_diffcommit(diffset=diffset, commit_id='r2',
                                       parent_id='r1'),
            ]

            filediffs = [
                self.create_filediff(diffset=diffset, commit=commits[0],
                                     source_file='/foo', dest_file='/foo'),
                self.create_filediff(diffset=diffset, commit=commits[1],
                                     source_file='/bar', dest_file='/bar'),
            ]
            review = self.create_review(review_request, user=self.user)

            rsp = self.api_post(
                get_review_diff_comment_list_url(review),
                {
                    'filediff_id': filediffs[1].pk,
                    'base_filediff_id': filediffs[0].pk,
                    'issue_opened': True,
                    'first_line': 1,
                    'num_lines': 5,
                    'text': 'foo',
                },
                expected_status=400)

            self.assertEqual(rsp['stat'], 'fail')
            self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
            self.assertEqual(rsp['fields'], {
                'base_filediff_id': [
                    'This is not a valid base filediff ID.',
                ],
            })

        self.assertTrue(FileDiff.get_ancestors.called)

    @webapi_test_template
    def test_post_with_base_filediff_ancestor(self):
        """Testing the POST <URL> API with base_filediff_id belonging to
        the FileDiff's set of ancestors
        """
        self.spy_on(FileDiff.get_ancestors,
                    owner=FileDiff)

        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                create_with_history=True,
                publish=True)

            diffset = self.create_diffset(review_request)

            commits = [
                self.create_diffcommit(diffset=diffset, commit_id='r1',
                                       parent_id='r0'),
                self.create_diffcommit(diffset=diffset, commit_id='r2',
                                       parent_id='r1'),
            ]

            filediffs = [
                self.create_filediff(diffset=diffset, commit=commits[0],
                                     source_revision='123', dest_detail='124'),
                self.create_filediff(diffset=diffset, commit=commits[1],
                                     source_revision='124', dest_detail='125'),
            ]
            review = self.create_review(review_request, user=self.user)

            rsp = self.api_post(
                get_review_diff_comment_list_url(review),
                {
                    'filediff_id': filediffs[1].pk,
                    'base_filediff_id': filediffs[0].pk,
                    'issue_opened': True,
                    'first_line': 1,
                    'num_lines': 5,
                    'text': 'foo',
                },
                expected_mimetype=review_diff_comment_item_mimetype)

            self.assertEqual(rsp['stat'], 'ok')
            self.assertIn('diff_comment', rsp)

            item_rsp = rsp['diff_comment']
            comment = Comment.objects.get(pk=item_rsp['id'])

            self.compare_item(item_rsp, comment)

        self.assertTrue(FileDiff.get_ancestors.called)

    @webapi_test_template
    def test_post_with_draft_filediff(self):
        """Testing the POST <URL> API with a draft FileDiff"""
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user)
        diffset = self.create_diffset(review_request, draft=True)
        filediff = self.create_filediff(diffset)
        review = self.create_review(review_request, user=self.user)

        diff_comment_text = 'Test diff comment'

        rsp = self.api_post(
            get_review_diff_comment_list_url(review),
            {
                'filediff_id': filediff.pk,
                'first_line': 1,
                'num_lines': 5,
                'text': diff_comment_text,
            },
            expected_mimetype=review_diff_comment_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('diff_comment', rsp)
        self.assertEqual(rsp['diff_comment']['text'], diff_comment_text)

class ResourceItemTests(CommentItemMixin, ReviewRequestChildItemMixin,
                        BaseResourceTestCase, metaclass=BasicTestsMetaclass):
    """Testing the ReviewDiffCommentResource item APIs."""
    fixtures = ['test_users', 'test_scmtools']
    sample_api_url = 'review-requests/<id>/reviews/<id>/diff-comments/'
    resource = resources.review_diff_comment

    compare_item = _compare_item

    def setup_review_request_child_test(self, review_request):
        if not review_request.repository_id:
            # The group tests don't create a repository by default.
            review_request.repository = self.create_repository()
            review_request.save()

        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)
        review = self.create_review(review_request, publish=True)
        comment = self.create_diff_comment(review, filediff)

        return (get_review_diff_comment_item_url(review, comment.pk),
                review_diff_comment_item_mimetype)

    def setup_http_not_allowed_list_test(self, user):
        review_request = self.create_review_request(
            create_repository=True,
            submitter=user,
            publish=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)
        review = self.create_review(review_request, user=user)
        comment = self.create_diff_comment(review, filediff)

        return get_review_diff_comment_item_url(review, comment.pk)

    #
    # HTTP DELETE tests
    #

    def setup_basic_delete_test(self, user, with_local_site, local_site_name):
        review_request = self.create_review_request(
            create_repository=True,
            with_local_site=with_local_site,
            submitter=user,
            publish=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)
        review = self.create_review(review_request, user=user)
        comment = self.create_diff_comment(review, filediff)

        return (get_review_diff_comment_item_url(review, comment.pk,
                                                 local_site_name),
                [comment, review])

    def check_delete_result(self, user, comment, review):
        self.assertNotIn(comment, review.comments.all())

    def test_delete_with_interdiff(self):
        """Testing the
        DELETE review-requests/<id>/reviews/<id>/diff-comments/<id>/ API
        """
        comment_text = "This is a test comment."

        comment, review_request, review, interdiff_revision = \
            self._common_post_interdiff_comments(comment_text)

        self.api_delete(get_review_diff_comment_item_url(review, comment.pk))

        rsp = self.api_get(get_review_diff_comment_list_url(review),
                           expected_mimetype=review_diff_comment_list_mimetype)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('diff_comments', rsp)
        self.assertEqual(len(rsp['diff_comments']), 0)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name):
        review_request = self.create_review_request(
            create_repository=True,
            with_local_site=with_local_site,
            submitter=user,
            publish=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)
        review = self.create_review(review_request, user=user)
        comment = self.create_diff_comment(review, filediff)

        return (get_review_diff_comment_item_url(review, comment.pk,
                                                 local_site_name),
                review_diff_comment_item_mimetype,
                comment)

    def test_get_not_modified(self):
        """Testing the
        GET review-requests/<id>/reviews/<id>/diff-comments/<id>/ API
        with Not Modified response
        """
        review = self._create_diff_review()
        comment = Comment.objects.all()[0]

        self._testHttpCaching(
            get_review_diff_comment_item_url(review, comment.id),
            check_etags=True)

    @webapi_test_template
    def test_get_with_dvcs_disabled(self):
        """Testing the GET <URL> API when DVCS feature disabled"""
        with override_feature_check(dvcs_feature.feature_id, enabled=False):
            review_request = self.create_review_request(
                create_repository=True,
                publish=True)

            diffset = self.create_diffset(review_request)
            commit = self.create_diffcommit(diffset=diffset)

            filediff = self.create_filediff(diffset=diffset, commit=commit)

            review = self.create_review(review_request, user=self.user)
            comment = self.create_diff_comment(review, filediff)

            rsp = self.api_get(
                get_review_diff_comment_item_url(review, comment.pk),
                expected_mimetype=review_diff_comment_item_mimetype)

            self.assertEqual(rsp['stat'], 'ok')
            self.assertIn('diff_comment', rsp)

            item_rsp = rsp['diff_comment']
            self.compare_item(item_rsp, comment)

    @webapi_test_template
    def test_get_with_dvcs_enabled(self):
        """Testing the GET <URL> API when DVCS feature enabled"""
        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                publish=True)

            diffset = self.create_diffset(review_request)
            commit = self.create_diffcommit(diffset=diffset)

            filediff = self.create_filediff(diffset=diffset, commit=commit)

            review = self.create_review(review_request, user=self.user)
            comment = self.create_diff_comment(review, filediff)

            rsp = self.api_get(
                get_review_diff_comment_item_url(review, comment.pk),
                expected_mimetype=review_diff_comment_item_mimetype)

            self.assertEqual(rsp['stat'], 'ok')
            self.assertIn('diff_comment', rsp)

            item_rsp = rsp['diff_comment']
            self.compare_item(item_rsp, comment)

    #
    # HTTP PUT tests
    #

    def setup_basic_put_test(self, user, with_local_site, local_site_name,
                             put_valid_data):
        review_request = self.create_review_request(
            create_repository=True,
            with_local_site=with_local_site,
            submitter=user,
            publish=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)
        review = self.create_review(review_request, user=user)
        comment = self.create_diff_comment(review, filediff)

        return (get_review_diff_comment_item_url(review, comment.pk,
                                                 local_site_name),
                review_diff_comment_item_mimetype,
                {'text': 'My new text'},
                comment,
                [])

    def check_put_result(self, user, item_rsp, comment, *args):
        self.assertEqual(item_rsp['id'], comment.pk)
        self.assertEqual(item_rsp['text'], 'My new text')
        self.assertEqual(item_rsp['text_type'], 'plain')
        self.compare_item(item_rsp, Comment.objects.get(pk=comment.pk))

    def test_put_with_issue(self):
        """Testing the
        PUT review-requests/<id>/reviews/<id>/diff-comments/<id>/ API,
        removing issue_opened
        """
        comment, review, review_request = self._create_diff_review_with_issue()

        rsp = self.api_put(
            get_review_diff_comment_item_url(review, comment.id),
            {'issue_opened': False},
            expected_mimetype=review_diff_comment_item_mimetype)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertFalse(rsp['diff_comment']['issue_opened'])

    def test_put_issue_status_before_publish(self):
        """Testing the
        PUT review-requests/<id>/reviews/<id>/diff-comments/<id>/ API
        with an issue, before review is published
        """
        comment, review, review_request = self._create_diff_review_with_issue()

        # The issue_status should not be able to be changed while the review is
        # unpublished.
        rsp = self.api_put(
            get_review_diff_comment_item_url(review, comment.id),
            {'issue_status': 'resolved'},
            expected_mimetype=review_diff_comment_item_mimetype)
        self.assertEqual(rsp['stat'], 'ok')

        # The issue_status should still be "open"
        self.assertEqual(rsp['diff_comment']['issue_status'], 'open')

    def test_put_issue_status_after_publish(self):
        """Testing the
        PUT review-requests/<id>/reviews/<id>/diff-comments/<id>/ API
        with an issue, after review is published
        """
        comment, review, review_request = self._create_diff_review_with_issue(
            publish=True)

        rsp = self.api_put(
            get_review_diff_comment_item_url(review, comment.id),
            {'issue_status': 'resolved'},
            expected_mimetype=review_diff_comment_item_mimetype)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['diff_comment']['issue_status'], 'resolved')

    def test_put_issue_status_by_issue_creator(self):
        """Testing the
        PUT review-requests/<id>/reviews/<id>/diff-comments/<id>/ API
        permissions for issue creator
        """
        comment, review, review_request = self._create_diff_review_with_issue(
            publish=True)

        # Change the owner of the review request so that it's not owned by
        # self.user.
        review_request.submitter = User.objects.get(username='doc')
        review_request.save()

        # The review/comment (and therefore issue) is still owned by self.user,
        # so we should be able to change the issue status.
        rsp = self.api_put(
            get_review_diff_comment_item_url(review, comment.id),
            {'issue_status': 'dropped'},
            expected_mimetype=review_diff_comment_item_mimetype)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['diff_comment']['issue_status'], 'dropped')

    def test_put_issue_status_by_uninvolved_user(self):
        """Testing the
        PUT review-requests/<id>/reviews/<id>/diff-comments/<id>/ API
        permissions for an uninvolved user
        """
        comment, review, review_request = self._create_diff_review_with_issue(
            publish=True)

        # Change the owner of the review request and review so that they're
        # not owned by self.user.
        new_owner = User.objects.get(username='doc')
        review_request.submitter = new_owner
        review_request.save()
        review.user = new_owner
        review.save()

        rsp = self.api_put(
            get_review_diff_comment_item_url(review, comment.id),
            {'issue_status': 'dropped'},
            expected_status=403)
        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], PERMISSION_DENIED.code)

    def test_put_with_remove_issue_opened(self):
        """Testing the
        PUT review-requests/<id>/reviews/<id>/diff-comments/<id>/ API,
        removing the issue_opened state
        """
        comment, review, review_request = self._create_diff_review_with_issue()

        rsp = self.api_put(
            get_review_diff_comment_item_url(review, comment.id),
            {'issue_opened': False},
            expected_mimetype=review_diff_comment_item_mimetype)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(rsp['diff_comment']['issue_status'], '')



"""Add HostingServiceAccount.hosting_url.

Version Added:
    1.7.8
"""

from django_evolution.mutations import AddField
from django.db import models


MUTATIONS = [
    AddField('HostingServiceAccount', 'hosting_url', models.CharField,
             max_length=256, null=True)
]



"""Main interface for managing and looking up SSL/TLS certificates.

Any code that needs to look up SSL/TLS certificates or CA bundles for local
are expected to use :py:class:`CertificateManager`. There's a
:py:data:`cert_manager` instance available for use.

Version Added:
    6.0
"""

from __future__ import annotations

import logging
import ssl
from typing import TYPE_CHECKING, cast
from urllib.parse import urlparse

from django.core.cache import cache
from django.utils.functional import SimpleLazyObject
from django.utils.text import slugify
from django.utils.translation import gettext as _
from djblets.cache.backend import cache_memoize, make_cache_key
from djblets.siteconfig.models import SiteConfiguration
from djblets.util.filesystem import safe_join
from housekeeping import func_deprecated
from typing_extensions import Final, TypedDict

from reviewboard.admin.server import get_data_dir
from reviewboard.certs.cert import (Certificate,
                                    CertPurpose,
                                    CertificateFingerprints)
from reviewboard.certs.errors import (CertificateNotFoundError,
                                      InvalidCertificateError)
from reviewboard.certs.storage import cert_storage_backend_registry
from reviewboard.deprecation import RemovedInReviewBoard10_0Warning

if TYPE_CHECKING:
    from typelets.funcs import KwargsDict
    from typelets.django.json import SerializableDjangoJSONDictImmutable
    from typing_extensions import NotRequired, TypeAlias

    from reviewboard.certs.cert import CertificateBundle
    from reviewboard.certs.storage.base import (
        BaseCertificateStorageBackend,
        BaseStoredCertificate,
        BaseStoredCertificateBundle,
        BaseStoredCertificateFingerprints,
    )
    from reviewboard.site.models import LocalSite

    _CertStorageBackend: TypeAlias = BaseCertificateStorageBackend[
        BaseStoredCertificate,
        BaseStoredCertificateBundle,
        BaseStoredCertificateFingerprints,
    ]


logger = logging.getLogger(__name__)


class CertificateFilePaths(TypedDict):
    """A dictionary of file paths for a certificate and private key.

    Version Added:
        6.0
    """

    #: The absolute path to a certificate file.
    cert_file: str

    #: The absolute path to a private key file.
    #:
    #: This is only included if a private key is present.
    key_file: NotRequired[str]


class CertificateManager:
    """A manager for looking up and working with SSL/TLS certificates.

    The certificate manager is used to perform common certificate management
    and lookup operations and to mark certificates as verified. It's the
    primary interface between communication channels and stored certificates.

    Any code that communicates with an external server that may be protected
    by a TLS/SSL certificate is expected to use the certificate manager to
    retrieve fingerprints, certificate data, certificate paths, or CA bundles
    and pass them to the communication code, in order to ensure that
    certificates can be verified.

    Access to the underlying :py:attr:`storage_backend` is available, for
    code that needs to more closely manage the storage of certificate-related
    information.

    Additional storage backends can be registered and configured, providing
    other options for certificate storage management.

    To set a custom storage backend, set the ``certs_storage_backend``
    setting in the Site Configuration and specify the registered ID of the
    storage backend.

    Version Added:
        6.0
    """

    #: The ID of the default storage backend.
    DEFAULT_STORAGE_ID: Final[str] = 'file'

    ######################
    # Instance variables #
    ######################

    #: A loaded instance of the current storage backend.
    _storage_backend: _CertStorageBackend | None

    def __init__(self) -> None:
        """Initialize the certificate manager."""
        self._storage_backend = None

    @property
    def storage_backend(self) -> _CertStorageBackend:
        """The current configured storage backend.

        If the storage backend is not loaded, or the last-accessed backend
        is no longer the configured backend, then a new one will be loaded and
        returned.

        If a new storage backend cannot be loaded, then the default file-based
        storage backend will be used.

        Consumers are encouraged to assign the results to a local variable
        before use, rather than repeatedly calling this. However, they should
        not hold onto an instance beyond the lifetime of a HTTP request.

        Type:
            reviewboard.certs.storage.base.BaseCertificateStorageBackend
        """
        siteconfig = SiteConfiguration.objects.get_current()
        configured_backend_id = siteconfig.get('certs_storage_backend',
                                               default=self.DEFAULT_STORAGE_ID)

        if not isinstance(configured_backend_id, str):
            logger.error('Site configuration key "certs_storage_backend" '
                         'has a non-string value (%r). Falling back to '
                         '"%s".',
                         configured_backend_id, self.DEFAULT_STORAGE_ID)
            configured_backend_id = self.DEFAULT_STORAGE_ID

        backend = self._storage_backend

        if backend is None or configured_backend_id != backend.backend_id:
            # We need to set and return a new storage backend. Either we
            # didn't have one loaded before or it changed due to a
            # configuration change.
            backend_cls = cert_storage_backend_registry.get_backend(
                configured_backend_id)

            if backend_cls is None:
                logger.error('Unable to load SSL/TLS certificate storage '
                             'backend "%s". Falling back to file-based '
                             'storage.',
                             configured_backend_id)

                backend_cls = cert_storage_backend_registry.get_backend(
                    self.DEFAULT_STORAGE_ID)
                assert backend_cls is not None

            assert backend_cls.backend_id
            storage_path = safe_join(get_data_dir(), 'rb-certs',
                                     slugify(backend_cls.backend_id))

            backend = backend_cls(storage_path=storage_path)
            self._storage_backend = backend

        return backend

    def add_ca_bundle(
        self,
        bundle: CertificateBundle,
        *,
        local_site: (LocalSite | None) = None,
    ) -> BaseStoredCertificateBundle:
        """Add a root CA bundle to storage.

        The certificate bundle's name along with ``local_site`` are considered
        unique within the storage backend. If there's an existing bundle with
        this information, it will be replaced.

        Args:
            bundle (reviewboard.certs.cert.CertificateBundle):
                The bundle to add.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

        Returns:
            reviewboard.certs.storage.base.BaseStoredCertificateBundle:
            The stored data for the certificate bundle.
        """
        return self.storage_backend.add_ca_bundle(bundle,
                                                  local_site=local_site)

    def delete_ca_bundle(
        self,
        *,
        name: str,
        local_site: (LocalSite | None) = None,
    ) -> None:
        """Delete a root CA bundle from storage.

        Args:
            name (str):
                The unique name of the CA bundle in storage.

                This is in :term:`slug` format.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

        Raises:
            reviewboard.cert.errors.CertificateNotFoundError:
                The CA bundle was not found.

            reviewboard.cert.errors.CertificateStorageError:
                There was an error deleting this CA bundle.
        """
        self.storage_backend.delete_ca_bundle(name=name,
                                              local_site=local_site)

    def get_ca_bundle(
        self,
        *,
        name: str,
        local_site: (LocalSite | None) = None,
    ) -> CertificateBundle | None:
        """Return a CA bundle.

        Args:
            name (str):
                The unique name of the CA bundle in storage.

                This is in :term:`slug` format.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the bundle.

                If ``None``, the global site will be used.

        Returns:
            CertificateBundle:
            The resulting certificate bundle.
        """
        stored_bundle = self.storage_backend.get_stored_ca_bundle(
            name=name,
            local_site=local_site)

        if not stored_bundle:
            return None

        return stored_bundle.bundle

    def get_ca_bundles_dir(
        self,
        *,
        local_site: (LocalSite | None) = None,
    ) -> str:
        """Return a path containing all CA bundle files.

        Args:
            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the bundles.

                If ``None``, the global site will be used.

        Returns:
            str:
            The path to the bundles directory.
        """
        return self.storage_backend.get_ca_bundles_dir(local_site=local_site)

    def add_certificate(
        self,
        certificate: Certificate,
        *,
        local_site: (LocalSite | None) = None,
    ) -> BaseStoredCertificate:
        """Add a certificate to storage.

        Once added, the certificate's fingerprints will also be marked as
        verified.

        Args:
            certificate (reviewboard.certs.cert.Certificate):
                The certificate to add.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that will own the certificate.

                If ``None``, the global site will be used.

        Returns:
            reviewboard.certs.storage.base.BaseStoredCertificate:
            The stored data for the certificate.

        Raises:
            reviewboard.certs.errors.InvalidCertificateError:
                The provided certificate is invalid and can't be stored.
        """
        try:
            certificate.x509_cert
        except Exception:
            raise InvalidCertificateError(
                _("The provided certificate data is invalid or corrupt and "
                  "can't be stored."))

        assert certificate.fingerprints is not None

        stored_cert = self.storage_backend.add_certificate(
            certificate=certificate,
            local_site=local_site)

        if certificate.purpose == CertPurpose.TRUST:
            # Make sure this is marked as verified.
            self.mark_certificate_verified(
                hostname=certificate.hostname,
                port=certificate.port,
                local_site=local_site,
                fingerprints=certificate.fingerprints,
            )

        return stored_cert

    def delete_certificate(
        self,
        *,
        hostname: str,
        port: int,
        local_site: (LocalSite | None) = None,
        purpose: CertPurpose = CertPurpose.TRUST,
    ) -> None:
        """Delete a certificate from storage.

        If deleting a certificate used to trust a server, then the
        certificate's fingerprints in storage will be removed as well,
        preventing them from being used for verification.

        Version Added:
            8.0:
            Added the ``purpose`` argument.

        Args:
            hostname (str):
                The hostname of the certificate to delete.

            port (int):
                The port that served the certificate.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

            purpose (reviewboard.certs.cert.CertPurpose, optional):
                The purpose of the certificate to delete.

                This indicates whether the certificate to match is used
                to trust a remote server or authenticate Review Board.

                By default, this matches only certificates used to trust a
                server.

                Version Added:
                    8.0
        """
        self.storage_backend.delete_certificate(
            hostname=hostname,
            port=port,
            local_site=local_site,
            purpose=purpose,
        )

        if purpose == CertPurpose.TRUST:
            # Remove any verification information for this certificate.
            self.remove_certificate_verification(
                hostname=hostname,
                port=port,
                local_site=local_site)

    def get_certificate(
        self,
        *,
        hostname: str,
        port: int,
        local_site: (LocalSite | None) = None,
        purpose: CertPurpose = CertPurpose.TRUST,
    ) -> Certificate | None:
        """Return a certificate for the given host, port, and purpose.

        By default, certificates used for trust will be returned. If a
        caller needs a certificate used for client authentication, that
        must be requested explicitly.

        Version Added:
            8.0:
            Added the ``purpose`` argument.

        Args:
            hostname (str):
                The hostname matching the certificate.

            port (int):
                The port serving the certificate.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

            purpose (reviewboard.certs.cert.CertPurpose, optional):
                The purpose of the certificate to fetch.

                This indicates whether the returned certificate should be used
                to trust a remote server or authenticate Review Board.

                By default, this returns a certificate used to trust a
                server.

                Version Added:
                    8.0

        Returns:
            reviewboard.certs.storage.base.BaseStoredCertificate:
            The resulting certificate, or ``None`` if not found.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error retrieving this certificate.
        """
        stored_cert = self.storage_backend.get_stored_certificate(
            hostname=hostname,
            port=port,
            local_site=local_site,
            purpose=purpose,
        )

        if not stored_cert:
            return None

        return stored_cert.certificate

    def get_certificate_file_paths(
        self,
        *,
        hostname: str,
        port: int,
        local_site: (LocalSite | None) = None,
        purpose: CertPurpose = CertPurpose.TRUST,
    ) -> CertificateFilePaths | None:
        """Return file paths for a certificate.

        Version Added:
            8.0:
            Added the ``purpose`` argument.

        Args:
            hostname (str):
                The hostname matching the certificate.

            port (int):
                The port serving the certificate.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

            purpose (reviewboard.certs.cert.CertPurpose, optional):
                The purpose of the certificate.

                This indicates whether the certificate should be used
                to trust a remote server or authenticate Review Board.

                By default, only certificates used for trust are returned.

                Version Added:
                    8.0

        Returns:
            CertificateFilePaths:
            The dictionary of file paths, or ``None`` if no paths are found.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error retrieving this certificate.
        """
        # TODO: Cache this in the future, to avoid repeated lookups.
        #       We will need to invalidate on cert updates.
        stored_cert = self.storage_backend.get_stored_certificate(
            hostname=hostname,
            port=port,
            local_site=local_site,
            purpose=purpose,
        )

        if not stored_cert:
            return None

        result: CertificateFilePaths = {
            'cert_file': stored_cert.get_cert_file_path(),
        }

        key_file_path = stored_cert.get_key_file_path()

        if key_file_path:
            result['key_file'] = key_file_path

        return result

    def mark_certificate_verified(
        self,
        *,
        hostname: str,
        port: int,
        fingerprints: CertificateFingerprints,
        local_site: (LocalSite | None) = None,
    ) -> None:
        """Mark a certificate as verified.

        The verification information will be stored in the certificate storage
        backend. It will also be cached in the cache backend for fast
        retrieval and subsequent verification.

        Args:
            hostname (str):
                The hostname that served the certificate.

            port (int):
                The port that served the certificate.

            fingerprints (reviewboard.certs.cert.CertificateFingerprints):
                The fingerprints to store in the verification storage.

                It's recommended to make this as thorough as possible, in
                order to more accurately match fingerprints that may come
                from different communication channels.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite to associate with the verification
                information.

                If ``None``, the global site will be used.

        Raises:
            reviewboard.certs.errors.InvalidCertificateError:
                The fingerprints were empty.
        """
        if fingerprints.is_empty():
            raise ValueError(
                'One or more SSL certificate fingerprints must be provided.'
            )

        # Store the fingerprints for verification in the backend.
        self.storage_backend.add_fingerprints(
            fingerprints,
            hostname=hostname,
            port=port,
            local_site=local_site)

        # Store this key in cache for efficient lookup.
        cache_key = self._build_fingerprints_cache_key(
            hostname=hostname,
            port=port,
            local_site=local_site)

        cache.set(make_cache_key(cache_key), fingerprints.to_json())

    def remove_certificate_verification(
        self,
        *,
        hostname: str,
        port: int,
        local_site: (LocalSite | None) = None,
    ) -> None:
        """Remove verification information for a certificate.

        Fingerprints for this certificate will no longer be returned as
        verified.

        Args:
            hostname (str):
                The hostname that served the certificate.

            port (int):
                The port that served the certificate.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite to associate with the verification
                information.

                If ``None``, the global site will be used.
        """
        # Store the fingerprints for verification in the backend.
        try:
            self.storage_backend.delete_fingerprints(
                hostname=hostname,
                port=port,
                local_site=local_site)
        except CertificateNotFoundError:
            # We can safely ignore this. The data wasn't found in storage,
            # which is the end goal anyway.
            pass

        # Remove any cached information about these fingerprints.
        cache.delete(make_cache_key(self._build_fingerprints_cache_key(
            hostname=hostname,
            port=port,
            local_site=local_site)))

    def get_verified_fingerprints(
        self,
        *,
        hostname: str,
        port: int,
        local_site: (LocalSite | None) = None,
    ) -> CertificateFingerprints | None:
        """Return certificate fingerprints from storage.

        Args:
            hostname (str):
                The hostname that served the certificate.

            port (int):
                The port that served the certificate.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite to associate with the verification
                information.

                If ``None``, the global site will be used.

        Returns:
            reviewboard.certs.cert.CertificateFingerprints:
            The resulting fingerprints, or ``None`` if not found in storage.
        """
        cache_key = self._build_fingerprints_cache_key(
            hostname=hostname,
            port=port,
            local_site=local_site)

        def _get_stored() -> SerializableDjangoJSONDictImmutable | None:
            stored_fingerprints = self.storage_backend.get_stored_fingerprints(
                hostname=hostname,
                port=port,
                local_site=local_site)

            if stored_fingerprints is None:
                return None

            return stored_fingerprints.fingerprints.to_json()

        data = cache_memoize(cache_key, _get_stored)

        if not data:
            return None

        return CertificateFingerprints.from_json(data)

    def is_certificate_verified(
        self,
        *,
        hostname: str,
        port: int,
        latest_fingerprints: CertificateFingerprints,
        local_site: (LocalSite | None) = None,
    ) -> bool:
        """Return whether a certificate is verified.

        This will determine verification based on the fingerprints. If one or
        more fingerprints match, and no fingerprints fail to match, then the
        certificate is considered verified.

        The result will be cached in the cache backend for fast retrieval and
        subsequent verification.

        Args:
            hostname (str):
                The hostname of the server to check for verification.

            port (int):
                The port of the server to check for verification.

            latest_fingerprints (reviewboard.certs.cert.
                                 CertificateFingerprints):
                The latest fingerprints from the server to check verified
                fingerprints against.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite the stored verification information
                would be associated with.

                If ``None``, the global site will be used.

        Returns:
            bool:
            ``True`` if the fingerprints have been verified. ``False`` if
            they have not.
        """
        verified_fingerprints = self.get_verified_fingerprints(
            hostname=hostname,
            port=port,
            local_site=local_site)

        return (verified_fingerprints is not None and
                verified_fingerprints.matches(latest_fingerprints))

    def build_ssl_context(
        self,
        *,
        hostname: str,
        port: int,
        local_site: (LocalSite | None) = None,
    ) -> ssl.SSLContext:
        """Return a configured SSL context for the given host.

        The resulting :py:class:`ssl.SSLContext` will be configured to use
        the CA bundles and certificates applicable to this host, port, and
        Local Site. This includes any certificates used for trust and any
        used for authentication.

        Args:
            hostname (str):
                The hostname of the server to connect to.

            port (int):
                The port of the server to connect to.

            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site that owns any stored certificates.

        Returns:
            ssl.SSLContext:
            The resulting SSL context.
        """
        # Set up an SSL context and load in any CA bundles.
        context = ssl.create_default_context()
        context.load_verify_locations(
            capath=self.get_ca_bundles_dir(local_site=local_site))

        # Load any certificates used to trust this host.
        trust_cert_paths = self.get_certificate_file_paths(
            hostname=hostname,
            port=port,
            purpose=CertPurpose.TRUST,
            local_site=local_site,
        )

        if trust_cert_paths:
            context.load_verify_locations(cafile=trust_cert_paths['cert_file'])

        # Load any certificates used to authenticate Review Board with the
        # server.
        client_cert_paths = self.get_certificate_file_paths(
            hostname=hostname,
            port=port,
            purpose=CertPurpose.CLIENT,
            local_site=local_site,
        )

        if client_cert_paths:
            if key_file := client_cert_paths.get('key_file'):
                context.load_cert_chain(
                    certfile=client_cert_paths['cert_file'],
                    keyfile=key_file,
                )
            else:
                logger.warning('Unable to locate key for SSL/TLS client '
                               'authentication certificate for %s:%s. The '
                               'certificate will be ignored.',
                               hostname, port)

        return context

    @func_deprecated(RemovedInReviewBoard10_0Warning)
    def build_urlopen_kwargs(
        self,
        *,
        url: str,
        local_site: (LocalSite | None) = None,
    ) -> KwargsDict:
        """Return SSL-related keyword arguments for a urlopen request.

        This will build keyword arguments that can be passed when calling
        :py:func:`urllib.request.urlopen`.

        If connecting to a HTTPS-based site, the keyword arguments will contain
        a pre-configured SSL context based on any stored CA bundles or
        certificates.

        If not connecting to a HTTPS-based site, this will be an empty
        dictionary.

        Deprecated:
            8.1:
            This is deprecated and will be removed in Review Board 10. Use
            :py:func:`reviewboard.certs.http.urlopen` or
            :py:class:`reviewboard.certs.http.
            CertificateVerificationHTTPSHandler` instead.

        Args:
            url (str):
                The URL to connect to.

            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site that owns any stored certificates.

        Returns:
            dict:
            The dictionary of keyword arguments.
        """
        parsed_url = urlparse(url)

        if parsed_url and parsed_url.hostname and parsed_url.scheme == 'https':
            return {
                'context': self.build_ssl_context(
                    hostname=parsed_url.hostname,
                    port=parsed_url.port or 443,
                    local_site=local_site),
            }

        return {}

    def _build_fingerprints_cache_key(
        self,
        *,
        hostname: str,
        port: int,
        local_site: LocalSite | None,
    ) -> str:
        """Return a cache key for verification.

        This key should be normalized before going into cache.

        Args:
            hostname (str):
                The hostname to include in the cache key.

            port (int):
                The port to include in the cache key.

            local_site (reviewboard.site.models.LocalSite):
                An optional LocalSite to include in the cache key.

        Returns:
            str:
            The resulting cache key.
        """
        backend_id = self.storage_backend.backend_id
        key = f'rb-ssl-fingerprints:{backend_id}'

        if local_site:
            key = f'{key}:{local_site.pk}'

        key = f'{key}:{hostname}:{port}'

        return key


#: The main certificate manager for Review Board.
#:
#: Version Added:
#:     6.0
cert_manager: Final[CertificateManager] = \
    cast(CertificateManager, SimpleLazyObject(CertificateManager))



#!/usr/bin/env python
"""Builds a Docker image for the current version of Review Board."""

import argparse
import os
import subprocess
import sys

docker_dir = os.path.dirname(__file__)
sys.path.insert(0, os.path.abspath(os.path.join(docker_dir, '..', '..')))
os.environ['DJANGO_SETTINGS_MODULE'] = 'reviewboard.settings'

from reviewboard import VERSION, get_package_version, is_release


IMAGE_NAME = 'beanbag/reviewboard'


if __name__ == '__main__':
    argparser = argparse.ArgumentParser(
        description='Build and optionally upload the Docker image.')
    argparser.add_argument(
        '--latest',
        dest='tag_latest',
        action='store_true',
        help='whether to tag with "latest" (should only be used if '
             'this is the very latest public version of Review Board.')
    argparser.add_argument(
        '--no-major',
        action='store_false',
        dest='tag_major',
        help='disable tagging the image with the "X.Y" major version tag')
    argparser.add_argument(
        '--load',
        action='store_true',
        help='load the image into Docker (this requires a single --platform).')
    argparser.add_argument(
        '--platform',
        action='append',
        help='specify an explicit platform to build.')
    argparser.add_argument(
        '--upload',
        action='store_true',
        help='upload the image after build')

    options = argparser.parse_args()

    package_version = get_package_version()
    major_version = '%s.%s' % VERSION[:2]
    image_version = package_version

    # If this is a development release, check if a built package has been
    # placed in the packages/ directory.
    if not is_release():
        package_version = '%s.dev0' % package_version
        package_path = os.path.join(docker_dir, 'packages',
                                    'reviewboard-%s-py3-none-any.whl'
                                    % package_version)

        if not os.path.exists(package_path):
            sys.stderr.write(
                'To build a Docker image for an in-development '
                'version of Review Board, you will\n'
                'need to build a development and place it at:\n'
                '\n'
                '%s\n'
                % package_path)
            sys.exit(1)

    platforms = options.platform or [
        'linux/amd64',
        'linux/arm64',
    ]

    if options.load and len(platforms) != 1:
        sys.stderr.write('--load requires a single --platform to be '
                         'specified.\n')
        sys.exit(1)

    if options.load and options.upload:
        sys.stderr.write('--load and --upload cannot both be specified.\n')
        sys.exit(1)

    tags = ['%s:%s' % (IMAGE_NAME, image_version)]

    if options.tag_major:
        tags.append('%s:%s' % (IMAGE_NAME, major_version))

    if options.tag_latest:
        tags.append(IMAGE_NAME)

    # Build the Docker command line to run.
    cmd = [
        'docker', 'buildx', 'build',
        '--platform', ','.join(platforms),
        '--build-arg', 'REVIEWBOARD_VERSION=%s' % package_version,
    ]

    if options.upload:
        cmd.append('--push')
    elif options.load:
        cmd.append('--load')

    for tag in tags:
        cmd += ['-t', tag]

    cmd.append('.')

    # Now build the image.
    p = subprocess.Popen(cmd,
                         stdin=sys.stdin,
                         stdout=sys.stdout,
                         stderr=sys.stderr,
                         cwd=docker_dir)

    if p.wait() != 0:
        sys.exit(1)
